"""
Created on 5 Mar 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://www.airnow.gov/
"""

import pytz
import re

from datetime import datetime as dt, timedelta as td, timezone as tz

from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class AQCSVDatetime(JSONable):
    """
    classdocs
    """

    @classmethod
    def construct_from_code(cls, code):
        try:
            match = re.match(r'(\d{4})(\d{2})(\d{2})T(\d{2})(\d{2})(([+\-])(\d{2})(\d{2}))?', code)
        except TypeError:
            raise ValueError(code)

        if match is None:
            raise ValueError(code)

        fields = match.groups()

        # datetime...
        year = int(fields[0])
        month = int(fields[1])
        day = int(fields[2])

        hour = int(fields[3])
        minute = int(fields[4])

        # no zone...
        if fields[5] is None:
            datetime = dt(year, month=month, day=day, hour=hour, minute=minute, tzinfo=pytz.timezone('Etc/UTC'))

            return AQCSVDatetime(datetime)

        # zone...
        zone_sign = -1 if fields[6] == '-' else 1
        zone_hours = int(fields[7])
        zone_mins = int(fields[8])

        zone_offset = zone_sign * td(hours=zone_hours, minutes=zone_mins)
        reporting_zone = tz(zone_offset)

        datetime = dt(year, month=month, day=day, hour=hour, minute=minute, tzinfo=reporting_zone)

        return AQCSVDatetime(datetime, reporting_zone)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, datetime, reporting_zone=None):
        """
        Constructor
        """
        self.__datetime = datetime                              # datetime
        self.__reporting_zone = reporting_zone                  # timezone


    def __hash__(self):
        return hash(self.localised())


    def __eq__(self, other):
        try:
            return self.datetime == other.datetime

        except (TypeError, AttributeError):
            return False


    def __ge__(self, other):
        return self.datetime >= other.datetime


    def __gt__(self, other):
        return self.datetime > other.datetime


    def __le__(self, other):
        return self.datetime <= other.datetime


    def __lt__(self, other):
        return self.datetime < other.datetime


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        if self.reporting_zone is None:
            datetime = self.datetime.astimezone(pytz.timezone('Etc/UTC'))
            return datetime.strftime("%Y%m%dT%H%M")

        datetime = self.localised()

        return datetime.strftime("%Y%m%dT%H%M%z")


    # ----------------------------------------------------------------------------------------------------------------

    def localised(self):
        return self.datetime.astimezone(self.__reporting_zone)


    def filename_prefix(self):
        datetime = self.datetime.astimezone(pytz.timezone('Etc/UTC'))
        return datetime.strftime("%Y%m%d%H%M")


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def datetime(self):
        return self.__datetime


    @property
    def reporting_zone(self):
        return self.__reporting_zone


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "AQCSVDatetime:{datetime:%s, reporting_zone:%s}" % (self.datetime, self.reporting_zone)



"""
Created on 24 Apr 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

Alphasense A4 electrochemical sensors
"""

from scs_core.gas.a4.a4 import A4
from scs_core.gas.a4.a4_temp_comp import A4TempComp


# --------------------------------------------------------------------------------------------------------------------

A4TempComp.init()       # must be initialised before sensors
A4.init()



"""
Created on 17 Jun 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

Alert example (recurring):
{"id": 85, "description": "test", "topic": "south-coast-science-demo/brighton-urban/loc/1/particulates",
"field": "exg.val.pm2p5", "lower-threshold": null, "upper-threshold": 20.0, "alert-on-none": false,
"aggregation-period": {"type": "recurring", "interval": 1, "units": "M", "timezone": "Europe/London"},
"contiguous-alerts": false, "json-message": false, "creator-email-address": "bruno.beloff@southcoastscience.com",
"to": "bruno.beloff@southcoastscience.com", "bcc-list": [], "suspended": false}

Alert example (diurnal):
{"id": 107, "description": "be2-3-nightime-test", "topic": "south-coast-science-dev/development/loc/1/climate",
"field": "val.tmp", "lower-threshold": null, "upper-threshold": 10.0, "alert-on-none": false,
"aggregation-period": {"type": "diurnal", "start": "20:00:00", "end": "09:50:00", "timezone": "Europe/London"},
"contiguous-alerts": true, "json-message": false, "creator-email-address": "production@southcoastscience.com",
"to": "bruno.beloff@southcoastscience.com", "bcc-list": ["jade.page@southcoastscience.com"], "suspended": false}

AlertStatus example:
{"id": 77, "rec": "2021-09-07T11:40:00Z", "cause": "OK", "val": 589.6}

https://martinstapel.com/how-to-autoincrement-in-dynamo-db-if-you-really-need-to/
https://stackoverflow.com/questions/37072341/how-to-use-auto-increment-for-primary-key-id-in-dynamodb
"""

from collections import OrderedDict

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.datum import Datum
from scs_core.data.diurnal_period import DiurnalPeriod
from scs_core.data.json import JSONable, JSONify
from scs_core.data.recurring_period import RecurringPeriod
from scs_core.data.str import Str

from scs_core.email.email import Email, EmailRecipient


# --------------------------------------------------------------------------------------------------------------------

class AlertStatus(JSONable):
    """
    classdocs
    """

    BELOW_LOWER_THRESHOLD =     '<L'
    ABOVE_UPPER_THRESHOLD =     '>U'
    NULL_VALUE =                'NV'
    OK =                        'OK'

    @classmethod
    def causes(cls):
        return cls.BELOW_LOWER_THRESHOLD, cls.ABOVE_UPPER_THRESHOLD, cls.NULL_VALUE, cls.OK


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        id = jdict.get('id')

        rec = LocalizedDatetime.construct_from_iso8601(jdict.get('rec'))
        cause = jdict.get('cause')
        value = jdict.get('val')

        return cls(id, rec, cause, value)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, id, rec, cause, value):
        """
        Constructor
        """
        self.__id = Datum.int(id)                                   # int

        self.__rec = rec                                            # LocalizedDatetime
        self.__cause = cause                                        # string
        self.__value = Datum.float(value)                           # float


    def __lt__(self, other):
        if self.id < other.id:
            return True

        if self.id > other.id:
            return False

        if self.rec < other.rec:
            return True

        if self.rec > other.rec:
            return False

        return False


    # ----------------------------------------------------------------------------------------------------------------

    def is_excursion(self):
        return self.cause != 'OK'


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['id'] = self.id

        jdict['rec'] = self.rec
        jdict['cause'] = self.cause
        jdict['val'] = self.value

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def id(self):
        return self.__id


    @property
    def rec(self):
        return self.__rec


    @property
    def cause(self):
        return self.__cause


    @property
    def value(self):
        return self.__value


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "AlertStatus:{id:%s, rec:%s, cause:%s, value:%s}" %  \
               (self.id, self.rec, self.cause, self.value)


# --------------------------------------------------------------------------------------------------------------------

class AlertSpecification(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        id = jdict.get('id')

        description = jdict.get('description', '')

        topic = jdict.get('topic')
        field = jdict.get('field')

        lower_threshold = jdict.get('lower-threshold')
        upper_threshold = jdict.get('upper-threshold')
        alert_on_none = jdict.get('alert-on-none')

        agg_jdict = jdict.get('aggregation-period')
        is_diurnal = agg_jdict.get('type') == DiurnalPeriod.type()

        aggregation_period = DiurnalPeriod.construct_from_jdict(agg_jdict) if is_diurnal else \
            RecurringPeriod.construct_from_jdict(agg_jdict)

        test_interval = RecurringPeriod.construct_from_jdict(jdict.get('test-interval'))

        contiguous_alerts = jdict.get('contiguous-alerts', is_diurnal)          # backwards-compatibility

        creator_email_address = jdict.get('creator-email-address')

        suspended = jdict.get('suspended')

        to = EmailRecipient.construct_from_jdict(jdict.get('to'))

        bcc_dict = {}
        for bcc_jdict in jdict.get('bcc-list'):
            recipient = EmailRecipient.construct_from_jdict(bcc_jdict)
            bcc_dict[recipient.email_address] = recipient

        return cls(id, description, topic, field, lower_threshold, upper_threshold, alert_on_none,
                   aggregation_period, test_interval, contiguous_alerts,
                   creator_email_address, to, bcc_dict, suspended)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, id, description, topic, field, lower_threshold, upper_threshold, alert_on_none,
                 aggregation_period, test_interval, contiguous_alerts,
                 creator_email_address, to, bcc_dict, suspended):
        """
        Constructor
        """
        self.__id = Datum.int(id)                                       # int

        self.__description = description                                # string

        self.__topic = topic                                            # string topic
        self.__field = field                                            # string path

        self.__lower_threshold = Datum.float(lower_threshold)           # float                 updatable
        self.__upper_threshold = Datum.float(upper_threshold)           # float                 updatable
        self.__alert_on_none = bool(alert_on_none)                      # bool                  updatable

        self.__aggregation_period = aggregation_period                  # Period                updatable
        self.__test_interval = test_interval                            # RecurringPeriod       updatable

        self.__contiguous_alerts = bool(contiguous_alerts)              # bool

        self.__creator_email_address = creator_email_address            # string

        self.__to = to                                                  # EmailRecipient
        self.__bcc_dict = bcc_dict                                      # dict of email: DeviceMonitorRecipient
        self.__suspended = bool(suspended)                              # bool                  updatable


    def __lt__(self, other):
        if self.description.lower() > other.description.lower():
            return False

        if self.description.lower() < other.description.lower():
            return True

        if self.topic < other.topic:
            return True

        if self.topic > other.topic:
            return False

        if self.field < other.field:
            return True

        if self.field > other.field:
            return False

        if self.creator_email_address.lower() < other.creator_email_address.lower():
            return True

        if self.creator_email_address.lower() > other.creator_email_address.lower():
            return False

        if self.to.lower() < other.to.lower():
            return True

        if self.to.lower() > other.to.lower():
            return False

        return False


    def __contains__(self, email):
        recipient = EmailRecipient(email, None)

        if self.creator_email_address == email or self.to == recipient:
            return True

        return EmailRecipient(email, None) in self.bcc_list


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        if self.topic is None or self.field is None or self.aggregation_period is None or self.to is None:
            return False

        if not self.has_trigger():
            return False

        if not self.has_valid_thresholds():
            return False

        # if self.test_interval is not None and not self.test_interval < self.aggregation_period:
        #     return False

        return True


    def has_trigger(self):
        return self.alert_on_none or self.lower_threshold is not None or self.upper_threshold is not None


    def has_valid_thresholds(self):
        if self.lower_threshold is None or self.upper_threshold is None:
            return True

        return self.lower_threshold < self.upper_threshold


    def has_valid_aggregation_period(self):
        return self.aggregation_period is not None and self.aggregation_period.is_valid()


    def has_update_compatibility(self, other):
        if self.id != other.id or self.topic != other.topic or \
                self.creator_email_address != other.creator_email_address:      # or self.field != other.field
            return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    def status(self, value):
        if self.suspended:
            return None

        if value is None and self.alert_on_none:
            cause = AlertStatus.NULL_VALUE

        elif self.lower_threshold is not None and value is not None and value < self.lower_threshold:
            cause = AlertStatus.BELOW_LOWER_THRESHOLD

        elif self.upper_threshold is not None and value is not None and value > self.upper_threshold:
            cause = AlertStatus.ABOVE_UPPER_THRESHOLD

        else:
            cause = None

        return AlertStatus(self.id, LocalizedDatetime.now().utc(), cause, value)


    # ----------------------------------------------------------------------------------------------------------------

    def checkpoint(self):     # TODO: needs an offset if test_interval is used
        return self.aggregation_period.checkpoint()


    def cron(self, minutes_offset):
        return self.aggregation_period.cron(minutes_offset)


    def aws_cron(self, minutes_offset):
        return self.aggregation_period.aws_cron(minutes_offset)
        # return self.test_interval.cron(minutes_offset) if self.test_interval else \
        #     self.aggregation_period.cron(minutes_offset)


    def timedelta(self):
        return self.aggregation_period.timedelta()


    def start_datetime(self, origin: LocalizedDatetime):
        return self.aggregation_period.start_datetime(origin)


    def end_datetime(self, origin: LocalizedDatetime):
        return self.aggregation_period.end_datetime(origin)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['id'] = self.id

        jdict['description'] = self.description

        jdict['topic'] = self.topic
        jdict['field'] = self.field

        jdict['lower-threshold'] = self.lower_threshold
        jdict['upper-threshold'] = self.upper_threshold
        jdict['alert-on-none'] = self.alert_on_none

        jdict['aggregation-period'] = self.aggregation_period.as_json()

        if self.test_interval is not None:
            jdict['test-interval'] = self.test_interval.as_json()

        jdict['contiguous-alerts'] = self.contiguous_alerts

        if self.creator_email_address is not None:
            jdict['creator-email-address'] = self.creator_email_address

        jdict['to'] = self.to
        jdict['bcc-list'] = self.bcc_list
        jdict['suspended'] = self.suspended

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def id(self):
        return self.__id


    @property
    def description(self):
        return self.__description


    @property
    def topic(self):
        return self.__topic


    @property
    def field(self):
        return self.__field


    @property
    def lower_threshold(self):
        return self.__lower_threshold


    @property
    def upper_threshold(self):
        return self.__upper_threshold


    @property
    def alert_on_none(self):
        return self.__alert_on_none


    @property
    def aggregation_period(self):
        return self.__aggregation_period


    @property
    def test_interval(self):
        return self.__test_interval


    @property
    def contiguous_alerts(self):
        return self.__contiguous_alerts


    @property
    def creator_email_address(self):
        return self.__creator_email_address


    @creator_email_address.setter
    def creator_email_address(self, creator_email_address):
        self.__creator_email_address = creator_email_address


    @property
    def to(self):
        return self.__to


    @property
    def bcc_list(self):
        return sorted(self.__bcc_dict.values())


    @property
    def bcc_dict(self):
        return self.__bcc_dict


    @property
    def suspended(self):
        return self.__suspended


    # ----------------------------------------------------------------------------------------------------------------

    @id.setter
    def id(self, id):
        self.__id = id


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        bcclist = Str.collection(self.bcc_list)

        return "AlertSpecification:{id:%s, description:%s, topic:%s, field:%s, lower_threshold:%s, " \
               "upper_threshold:%s, alert_on_none:%s, aggregation_period:%s, test_interval:%s, " \
               "contiguous_alerts:%s, creator_email_address:%s, to:%s, bcc_list:%s, suspended:%s}" %  \
               (self.id, self.description, self.topic, self.field, self.lower_threshold,
                self.upper_threshold, self.alert_on_none, self.aggregation_period, self.test_interval,
                self.contiguous_alerts, self.creator_email_address, self.to, bcclist, self.suspended)


# --------------------------------------------------------------------------------------------------------------------

class AlertMessage(Email, JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        alert_specification = AlertSpecification.construct_from_jdict(jdict.get('specification'))
        alert_status = AlertStatus.construct_from_jdict(jdict.get('status'))

        return cls(alert_specification, alert_status)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, alert_specification: AlertSpecification, alert_status: AlertStatus):
        """
        Constructor
        """
        self.__alert_specification = alert_specification                    # AlertSpecification
        self.__alert_status = alert_status                                  # AlertStatus


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['specification'] = self.alert_specification
        jdict['status'] = self.alert_status

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def subject(self):
        return JSONify.dumps(self.alert_status)


    @property
    def body(self):
        return JSONify.dumps(self)


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def alert_specification(self):
        return self.__alert_specification


    @property
    def alert_status(self):
        return self.__alert_status


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "AlertMessage:{alert_specification:%s, alert_status:%s}" %  \
               (self.alert_specification, self.alert_status)



"""
Created on 15 Mar 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

A stub class for an NDIRConf that may be implemented elsewhere
"""

from collections import OrderedDict

from scs_core.data.json import PersistentJSONable


# --------------------------------------------------------------------------------------------------------------------

class NDIRConf(PersistentJSONable):
    """
    classdocs
    """

    __FILENAME = "ndir_conf.json"

    @classmethod
    def persistence_location(cls):
        return cls.conf_dir(), cls.__FILENAME


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        model = jdict.get('model')
        tally = jdict.get('tally')
        raw = jdict.get('raw', False)

        return cls(model, tally, raw)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, model, tally, raw=False):
        """
        Constructor
        """
        super().__init__()

        self.__model = model
        self.__tally = tally
        self.__raw = raw


    def __eq__(self, other):
        try:
            return self.model == other.model and self.tally == other.tally and self.raw == other.raw

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def ndir_monitor(self, interface, host):
        return None


    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def ndir(self, interface, host):
        return None


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['model'] = self.model
        jdict['tally'] = self.tally
        jdict['raw'] = self.raw

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def model(self):
        return self.__model


    @property
    def tally(self):
        return self.__tally


    @property
    def raw(self):
        return self.__raw


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "NDIRConf(core):{model:%s, tally:%s, raw:%s}" %  (self.model, self.tally, self.raw)



"""
Created on 16 Apr 2018

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import csv
import time

from scs_core.csv.csv_dict import CSVDict

from scs_core.data.datetime import LocalizedDatetime

from scs_core.sys.filesystem import Filesystem
from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

class CSVLogger(object):
    """
    classdocs
    """

    __MIN_FREE_SPACE = 10485760                         # 10MB
    __CHECK_INTERVAL = 4                                # normally once per day


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, host, log, delete_oldest, write_interval):
        manager = CSVSpaceManager(host, log, delete_oldest, cls.__MIN_FREE_SPACE, cls.__CHECK_INTERVAL)

        return cls(host, log, manager, write_interval)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, host, log, manager, write_interval):
        """
        Constructor
        """
        self.__host = host                              # PersistenceManager
        self.__log = log                                # CSVLog
        self.__manager = manager                        # CSVSpaceManager
        self.__write_interval = write_interval          # int

        self.__paths = None                             # array of string
        self.__file = None                              # file handle
        self.__latest_write = None                      # timestamp
        self.__writing_inhibited = False                # bool

        self.__buffer = []                              # array of CSVDict


    # ----------------------------------------------------------------------------------------------------------------

    def write(self, jstr):
        if self.writing_inhibited:
            return None

        if jstr is None or self.log is None:
            return None

        datum = CSVDict.construct_from_jstr(jstr)

        if datum is None:
            return None

        # direct write...
        if not self.write_interval:
            self.__write(datum)
            self.__file.flush()
            return self.log.file_path()

        # interval write...
        now = time.time()

        if self.__latest_write is None:
            self.__latest_write = now

        interval = now - self.__latest_write

        # append to buffer...
        if interval < self.write_interval:
            self.__buffer.append(datum)
            return self.log.file_path()

        self.__latest_write = now

        # deferred write...
        for datum in self.__buffer:
            self.__write(datum)

        self.__file.flush()

        self.__buffer = []

        return self.log.file_path()


    def close(self):
        if self.__file is None:
            return

        self.__file.close()
        self.__file = None


    # ----------------------------------------------------------------------------------------------------------------

    def __write(self, datum):
        # first run...
        if not self.__file:
            if self.log.tag is None and 'tag' in datum.collection:
                self.log.tag = datum.collection['tag']

            self.__open_file()

        # start log for new day...
        if not self.log.in_timeline(LocalizedDatetime.now().utc()):
            self.close()
            self.__open_file()

        # write header...
        if not self.__paths:
            self.__paths = datum.paths()
            self.__writer.writerow(self.__paths)

        # write row...
        self.__writer.writerow(datum.row(self.__paths))


    def __open_file(self):
        self.log.timeline_start = LocalizedDatetime.now().utc()

        # check...
        if not self.__manager.clear_space():
            self.writing_inhibited = True

        self.log.mkdir()

        # file...
        self.__file = open(self.log.file_path(), "w")
        self.__writer = csv.writer(self.__file, quoting=csv.QUOTE_MINIMAL)

        # header...
        if self.__paths:
            self.__writer.writerow(self.__paths)


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def log(self):
        return self.__log


    @property
    def write_interval(self):
        return self.__write_interval


    @property
    def writing_inhibited(self):
        return self.__writing_inhibited


    @writing_inhibited.setter
    def writing_inhibited(self, writing_inhibited):
        self.__writing_inhibited = writing_inhibited


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "CSVLogger:{log:%s, manager:%s, write_interval:%s, writing_inhibited:%s}" % \
               (self.log, self.__manager, self.write_interval, self.writing_inhibited)


# --------------------------------------------------------------------------------------------------------------------

class CSVSpaceManager(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, host, log, delete_oldest, min_free_space, check_interval):
        self.__logger = Logging.getLogger()

        self.__host = host                                          # PersistenceManager
        self.__log = log                                            # CSVLog
        self.__delete_oldest = delete_oldest                        # bool
        self.__min_free_space = min_free_space                      # int
        self.__check_interval = check_interval                      # int

        self.__check_count = check_interval


    # ----------------------------------------------------------------------------------------------------------------

    def clear_space(self):
        self.__check_count += 1

        if self.__check_count < self.__check_interval:
            return True

        self.__check_count = 0
        return self.__clear_space()


    # ----------------------------------------------------------------------------------------------------------------

    def __clear_space(self):
        if self.__has_sufficient_space():
            return True

        # stop on no-delete...
        if not self.__delete_oldest:
            self.__logger.error("CSVSpaceManager.__clear_space: volume full.")
            return False

        # delete until enough free...
        while not self.__has_sufficient_space():
            success = self.__delete_oldest_log()

            if not success:
                self.__logger.error("CSVSpaceManager.__clear_space: delete failed.")
                return False

        return True


    def __has_sufficient_space(self):
        du = self.__host.disk_usage(self.__log.root_path)

        if du is None:
            raise OSError("has_sufficient_space: '%s' is not available." % self.__log.root_path)

        return du.free > self.__min_free_space


    def __delete_oldest_log(self):
        # walk the directories...
        containers = Filesystem.ls(self.__log.root_path)

        for container in containers:
            if not container.is_directory:
                continue

            # walk the files...
            files = Filesystem.ls(container.path())

            for file in files:
                if not file.is_directory and file.has_suffix('csv'):
                    self.__logger.error("CSVSpaceManager.__delete_oldest_log: deleting: %s" % file)

                    success = file.delete()

                    if not success:
                        return False

                    Filesystem.rmdir(container.path())          # remove empty directories

                    return True

        return False


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "CSVSpaceManager:{delete_oldest:%s, min_free_space:%s}" %  \
               (self.__delete_oldest, self.__min_free_space)



"""
Created on 11 Jan 2021

@author: Jade Page (jade.page@southcoastscience.com)
"""

from scs_core.aws.greengrass.aws_group import AWSGroup


# --------------------------------------------------------------------------------------------------------------------

class AWSGroupDeployer(object):
    """
    classdocs
    """

    BUILDING =      "Building"
    IN_PROGRESS =   "InProgress"
    SUCCESS =       "Success"
    FAILURE =       "Failure"

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, group_name, client):
        self.__group_name = group_name
        self.__client = client


    # ----------------------------------------------------------------------------------------------------------------

    def retrieve_deployment_info(self, client):
        aws_group_info = AWSGroup(self.__group_name, client)

        datum = aws_group_info.get_group_info_from_name()
        group_id = datum.node("GroupID")
        group_version_id = datum.node("GroupLatestVersionID")

        return group_id, group_version_id


    def deploy(self):
        group_id, group_version_id = self.retrieve_deployment_info(self.__client)

        response = self.__client.create_deployment(
            DeploymentType="NewDeployment",
            GroupId=group_id,
            GroupVersionId=group_version_id
        )

        return response


    def status(self, response):
        group_id, _ = self.retrieve_deployment_info(self.__client)

        w_response = self.__client.get_deployment_status(
            DeploymentId=response.get("DeploymentId"),
            GroupId=group_id
        )

        return w_response.get("DeploymentStatus")


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "AWSGroupDeployer:{group_name:%s, client:%s}" % (self.__group_name, self.__client)



"""
Created on 30 Dec 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

Course over ground and Ground speed
$xxVTG,cogt,T,cogm,M,knots,N,kph,K,pos_mode*cs

example sentence:
$GPVTG,77.52,T,,M,0.004,N,0.008,K,A*06

example values:
GPVTG:{cogt:None, cogm:None, knots:0.005, kph:0.010, pos_mode:D}
GPVTG:{cogt:None, cogm:None, knots:None, kph:None, pos_mode:N}

https://www.nmea.org
https://en.wikipedia.org/wiki/NMEA_0183
"""

from scs_core.position.nmea.nmea_sentence import NMEASentence


# --------------------------------------------------------------------------------------------------------------------

class GPVTG(NMEASentence):
    """
    classdocs
    """

    MESSAGE_IDS = ("$GNVTG", "$GPVTG")

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, r):
        if r.message_id not in cls.MESSAGE_IDS:
            raise TypeError("invalid sentence:%s" % r)

        cogt = r.float(1, 2)
        cogm = r.float(3, 2)

        knots = r.float(5, 3)
        kph = r.float(7, 3)

        pos_mode = r.str(9)

        return GPVTG(r.message_id, cogt, cogm, knots, kph, pos_mode)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, message_id, cogt, cogm, knots, kph, pos_mode):
        """
        Constructor
        """
        super().__init__(message_id)

        self.__cogt = cogt                      # float(2) - degrees course over ground
        self.__cogm = cogm                      # float(2) - degrees course over ground (magnetic)

        self.__knots = knots                    # float(3) - speed over ground (knots)
        self.__kph = kph                        # float(3) - speed over ground (kph)

        self.__pos_mode = pos_mode              # string


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def cogt(self):
        return self.__cogt


    @property
    def cogm(self):
        return self.__cogm


    @property
    def knots(self):
        return self.__knots


    @property
    def kph(self):
        return self.__kph


    @property
    def pos_mode(self):
        return self.__pos_mode


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "GPVTG:{source:%s, cogt:%s, cogm:%s, knots:%s, kph:%s, pos_mode:%s}" % \
                    (self.source, self.cogt, self.cogm, self.knots, self.kph, self.pos_mode)



"""
Created on 26 Nov 2018

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

from collections import OrderedDict
from urllib.parse import parse_qs, urlparse

from scs_core.aws.client.api_intercourse import APIResponse
from scs_core.aws.data.message import Message

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.datum import Datum
from scs_core.data.str import Str
from scs_core.data.timedelta import Timedelta


# --------------------------------------------------------------------------------------------------------------------

class TopicHistoryRequest(object):
    """
    classdocs
    """

    TOPIC = 'topic'
    START = 'startTime'
    END = 'endTime'
    PATH = 'path'
    FETCH_LAST_WRITTEN = 'fetchLastWritten'
    CHECKPOINT = 'checkpoint'
    INCLUDE_WRAPPER = 'includeWrapper'
    REC_ONLY = 'recOnly'
    MIN_MAX = 'minMax'
    EXCLUDE_REMAINDER = 'excludeRemainder'
    FETCH_LAST_WRITTEN_BEFORE = 'fetchLastWrittenBefore'
    BACKOFF_LIMIT = "backoffLimit"

    # ----------------------------------------------------------------------------------------------------------------

    __DAY = Timedelta(days=1)
    __WEEK = Timedelta(weeks=1)
    __MONTH = Timedelta(days=28)
    __YEAR = Timedelta(days=365)

    @classmethod
    def checkpoint_table(cls, start, end):
        delta = end - start

        if delta < cls.__DAY:
            return None  # raw data rate

        if delta < cls.__WEEK:
            return '**:/01:00'

        if delta < cls.__MONTH:
            return '**:/15:00'

        if delta < cls.__YEAR:
            return '**:00:00'

        return '/06:00:00'


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_qsp(cls, qsp):
        if not qsp:
            return None

        # compulsory...
        topic = qsp.get(cls.TOPIC)
        start = LocalizedDatetime.construct_from_iso8601(qsp.get(cls.START))
        end = LocalizedDatetime.construct_from_iso8601(qsp.get(cls.END))

        # optional...
        path = qsp.get(cls.PATH)
        fetch_last_written = Datum.is_true(qsp.get(cls.FETCH_LAST_WRITTEN))
        checkpoint = qsp.get(cls.CHECKPOINT)
        include_wrapper = Datum.is_true(qsp.get(cls.INCLUDE_WRAPPER))
        rec_only = Datum.is_true(qsp.get(cls.REC_ONLY))
        min_max = Datum.is_true(qsp.get(cls.MIN_MAX))
        exclude_remainder = Datum.is_true(qsp.get(cls.EXCLUDE_REMAINDER))
        fetch_last_written_before = Datum.is_true(qsp.get(cls.FETCH_LAST_WRITTEN_BEFORE))
        backoff_limit = qsp.get(cls.backoff_limit)

        if checkpoint and checkpoint.lower() == 'none':
            checkpoint = None

        return cls(topic, start, end, path, fetch_last_written, checkpoint, include_wrapper, rec_only,
                   min_max, exclude_remainder, fetch_last_written_before, backoff_limit)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, topic, start, end, path, fetch_last_written, checkpoint, include_wrapper, rec_only,
                 min_max, exclude_remainder, fetch_last_written_before, backoff_limit):
        """
        Constructor
        """
        self.__topic = topic                                                        # string
        self.__start = start                                                        # LocalizedDatetime
        self.__end = end                                                            # LocalizedDatetime

        self.__path = path                                                          # string
        self.__fetch_last_written = bool(fetch_last_written)                        # bool
        self.__checkpoint = checkpoint                                              # string
        self.__include_wrapper = bool(include_wrapper)                              # bool
        self.__rec_only = bool(rec_only)                                            # bool
        self.__min_max = bool(min_max)                                              # bool
        self.__exclude_remainder = bool(exclude_remainder)                          # bool
        self.__fetch_last_written_before = bool(fetch_last_written_before)          # bool
        self.__backoff_limit = backoff_limit                                        # int seconds


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        if self.topic is None or self.start is None or self.end is None:
            return False

        now = LocalizedDatetime.now()

        if self.start > now or self.end > now:
            return False

        if self.start > self.end:
            return False

        if self.include_wrapper and self.checkpoint is not None:
            return False

        if self.rec_only and self.fetch_last_written:
            return False

        if self.rec_only and self.path:
            return False

        if self.rec_only and self.min_max:
            return False

        if self.min_max and self.checkpoint is None:
            return False

        if self.exclude_remainder and self.checkpoint is None:
            return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    def next_params(self, start):
        return TopicHistoryRequest(self.topic, start, self.end, self.path, self.fetch_last_written, self.checkpoint,
                                   self.include_wrapper, self.rec_only, self.min_max, self.exclude_remainder,
                                   self.backoff_limit, self.fetch_last_written_before).params()


    def change_params(self, start, end):
        return TopicHistoryRequest(self.topic, start, end, self.path, self.fetch_last_written, self.checkpoint,
                                   self.include_wrapper, self.rec_only, self.min_max, self.exclude_remainder,
                                   self.backoff_limit, self.fetch_last_written_before)


    def params(self):
        params = {
            self.TOPIC: self.topic,
            self.START: None if self.start is None else self.start.utc().as_iso8601(include_millis=True),
            self.END: self.end.utc().as_iso8601(include_millis=True)
        }

        if self.path is not None:
            params[self.PATH] = self.path

        if self.fetch_last_written:
            params[self.FETCH_LAST_WRITTEN] = 'true'

        if self.checkpoint is not None:
            params[self.CHECKPOINT] = self.checkpoint

        if self.include_wrapper:
            params[self.INCLUDE_WRAPPER] = 'true'

        if self.rec_only:
            params[self.REC_ONLY] = 'true'

        if self.min_max:
            params[self.MIN_MAX] = 'true'

        if self.exclude_remainder:
            params[self.EXCLUDE_REMAINDER] = 'true'

        if self.fetch_last_written_before:
            params[self.FETCH_LAST_WRITTEN_BEFORE] = 'true'

        if self.backoff_limit:
            params[self.BACKOFF_LIMIT] = self.backoff_limit

        return params


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def checkpoint(self):
        if self.__checkpoint is None:
            return None

        if self.__checkpoint == 'auto':
            return self.checkpoint_table(self.start, self.end)

        return self.__checkpoint


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def topic(self):
        return self.__topic


    @property
    def start(self):
        return self.__start


    @property
    def end(self):
        return self.__end


    @property
    def path(self):
        return self.__path


    @property
    def fetch_last_written(self):
        return self.__fetch_last_written


    @property
    def include_wrapper(self):
        return self.__include_wrapper


    @property
    def rec_only(self):
        return self.__rec_only


    @property
    def min_max(self):
        return self.__min_max


    @property
    def exclude_remainder(self):
        return self.__exclude_remainder


    @property
    def fetch_last_written_before(self):
        return self.__fetch_last_written_before


    @property
    def backoff_limit(self):
        return self.__backoff_limit


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "TopicHistoryRequest:{topic:%s, start:%s, end:%s, path:%s, fetch_last_written:%s, checkpoint:%s, " \
               "include_wrapper:%s, rec_only:%s, min_max:%s, exclude_remainder:%s, " \
               "fetch_last_written_before:%s, backoff_limit:%s}" % \
               (self.topic, self.start, self.end, self.path, self.fetch_last_written, self.__checkpoint,
                self.include_wrapper, self.rec_only, self.min_max, self.exclude_remainder,
                self.fetch_last_written_before, self.backoff_limit)


# --------------------------------------------------------------------------------------------------------------------

class TopicHistoryResponse(APIResponse):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        fetched_last = jdict.get('fetchedLastWrittenData')
        interval = jdict.get('interval')

        items = []
        for msg_jdict in jdict.get('Items'):
            item = Message.construct_from_jdict(msg_jdict) if 'payload' in msg_jdict else msg_jdict
            items.append(item)

        next_url = jdict.get('next')

        return cls(fetched_last, interval, items, next_url)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, fetched_last, interval, items, next_url):
        """
        Constructor
        """
        self.__fetched_last = fetched_last                      # "Fetched last written data" flag
        self.__interval = interval                              # int

        self.__items = items                                    # list of Message

        self.__next_url = next_url                              # URL string


    def __len__(self):
        return len(self.items)


    # ----------------------------------------------------------------------------------------------------------------

    def next_params(self, params):
        return parse_qs(urlparse(self.next_url).query)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        if self.fetched_last is not None:
            jdict['fetchedLastWrittenData'] = self.fetched_last

        if self.items is not None:
            jdict['interval'] = None if self.interval is None else int(round(self.interval))

            jdict['Items'] = self.items
            jdict['itemCount'] = len(self.items)

        if self.next_url is not None:
            jdict['next'] = self.next_url

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def start(self):
        if not self.items:
            return None

        item = self.items[0]

        try:
            return item['rec']
        except TypeError:
            return item.payload['rec']


    def end(self):
        if not self.items:
            return None

        item = self.items[-1]

        try:
            return item['rec']
        except TypeError:
            return item.payload['rec']


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def fetched_last(self):
        return self.__fetched_last


    @property
    def interval(self):
        return self.__interval


    @property
    def items(self):
        return self.__items


    @property
    def next_url(self):
        return self.__next_url


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "TopicHistoryResponse:{fetched_last:%s, interval:%s, items:%s, next_url:%s}" % \
               (self.fetched_last, self.interval, Str.collection(self.items), self.next_url)



"""
Created on 25 Dec 2018

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example:
{"device": "scs-bgx-401", "topic": "south-coast-science-demo/brighton/loc/1/climate",
"lastSeenTime": "2020-10-23T08:52:20Z", "last_write": "2020-10-23T08:52:20Z",
"message": "{\"val\": {\"hmd\": 68.4, \"tmp\": 19.8, \"bar\": null}, \"rec\": \"2020-10-23T08:52:20Z\",
\"tag\": \"scs-bgx-401\"}"}
"""

import json

from collections import OrderedDict

from scs_core.aws.security.cognito_device import CognitoDeviceCredentials

from scs_core.data.array_dict import ArrayDict
from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONable
from scs_core.data.topic_path import TopicPath

from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

class Byline(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        device = jdict.get('device')
        topic = jdict.get('topic')

        pub = LocalizedDatetime.construct_from_iso8601(jdict.get('lastSeenTime'))
        rec = LocalizedDatetime.construct_from_iso8601(jdict.get('last_write'))

        try:
            jdict.get('message').keys()
            message = json.dumps(jdict.get('message'))      # web API - message is a dict

        except AttributeError:
            message = jdict.get('message')                  # this class - message is a string

        return cls(device, topic, pub, rec, message)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, device, topic, pub, rec, message):
        """
        Constructor
        """
        self.__device = device                      # string tag
        self.__topic = topic                        # string path

        self.__pub = pub                            # LocalizedDatetime
        self.__rec = rec                            # LocalizedDatetime

        self.__message = message                    # string


    def __lt__(self, other):
        # device...
        if self.__device < other.__device:
            return True

        if self.__device > other.__device:
            return False

        # topic...
        if self.__topic < other.__topic:
            return True

        if self.__topic > other.__topic:
            return False

        # rec...
        if self.__rec is None:
            return True

        if other.__rec is None:
            return False

        if self.__rec < other.__rec:
            return True

        return False


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['device'] = self.device
        jdict['topic'] = self.topic

        jdict['lastSeenTime'] = None if self.pub is None else self.pub.as_iso8601()
        jdict['last_write'] = None if self.rec is None else self.rec.as_iso8601()

        if self.message:
            jdict['message'] = self.message

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def device(self):
        return self.__device


    @property
    def topic(self):
        return self.__topic


    @property
    def pub(self):
        return self.__pub


    @property
    def rec(self):
        return self.__rec


    @property
    def message(self):
        return self.__message


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Byline:{device:%s, topic:%s, pub:%s, rec:%s, message:%s}" %  \
               (self.device, self.topic, self.pub, self.rec, self.message)


# --------------------------------------------------------------------------------------------------------------------

class BylineGroup(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, excluded=None, strict_tags=False, skeleton=False):
        if not jdict:
            return cls({}) if skeleton else None

        bylines = [Byline.construct_from_jdict(byline_jdict) for byline_jdict in jdict]

        return cls.construct(bylines, excluded=excluded, strict_tags=strict_tags)


    @classmethod
    def construct(cls, bylines, excluded=None, strict_tags=False):
        filtered_bylines = []

        for byline in bylines:
            if excluded is not None and byline.topic.endswith(excluded):
                continue

            if strict_tags and not CognitoDeviceCredentials.is_valid_tag(byline.device):
                continue

            filtered_bylines.append(byline)

        # device_bylines...
        device_bylines = ArrayDict([(byline.device, byline) for byline in sorted(filtered_bylines)])

        return cls(device_bylines)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, device_bylines):
        """
        Constructor
        """
        self._device_bylines = device_bylines                   # dict of device: Byline


    def __bool__(self):
        return len(self) > 0


    def __len__(self):
        return len(list(self.bylines))


    # ----------------------------------------------------------------------------------------------------------------

    def latest_byline(self, suffix=''):
        latest_byline = None

        for byline in self.bylines:
            if byline.topic.endswith(suffix) and (latest_byline is None or byline.rec > latest_byline.rec):
                latest_byline = byline

        return latest_byline


    def latest_topic(self, suffix=''):
        latest_rec = None
        topic = None

        for byline in self.bylines:
            if byline.topic.endswith(suffix) and (latest_rec is None or byline.rec > latest_rec):
                latest_rec = byline.rec
                topic = byline.topic

        return topic


    def latest_pub(self):
        if not self.bylines:
            return None

        return max([byline.pub for byline in self.bylines if byline.pub is not None], default=None)


    def latest_rec(self):
        if not self.bylines:
            return None

        return max([byline.rec for byline in self.bylines if byline.rec is not None], default=None)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        return tuple(byline.as_json() for byline in self.bylines)       # matches the structure of the API response


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def bylines(self):
        for bylines in self._device_bylines.values():
            for byline in bylines:
                yield byline


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return self.__class__.__name__ + ":{device_bylines:%s}" % self._device_bylines


# --------------------------------------------------------------------------------------------------------------------

class DeviceBylineGroup(BylineGroup):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, device_bylines):
        """
        Constructor
        """
        super().__init__(device_bylines)


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def device(self):
        for device in self._device_bylines.keys():
            return device                                       # return the first device tag

        return None


# --------------------------------------------------------------------------------------------------------------------

class TopicBylineGroup(BylineGroup):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, device_bylines):
        """
        Constructor
        """
        super().__init__(device_bylines)

        self.__logger = Logging.getLogger()


    # ----------------------------------------------------------------------------------------------------------------

    def device_byline_group(self, device):
        return DeviceBylineGroup({device: self._device_bylines[device]})


    def bylines_for_device(self, device):
        return self._device_bylines[device]


    def topic_roots(self):
        topic_roots = set()
        for device_tag, bylines in self._device_bylines.items():
            for byline in bylines:
                try:
                    topic_roots.add(TopicPath.construct(byline.rec, byline.topic).root())

                except ValueError:
                    self.__logger.debug("invalid topic: %s" % byline.topic)
                    continue

        return sorted(topic_roots)


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def devices(self):
        return list(self._device_bylines.keys())
        # return self._device_bylines.keys()



"""
Created on 9 Dec 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

The A4CalibratedDatum is designed to provide a model training data set that encapsulates the calibration of the
electrochemical sensor - cal_x_v is only relevant to sensors with NO2 cross-sensitivity.
"""

from collections import OrderedDict

from scs_core.data.datum import Datum
from scs_core.data.datetime import LocalizedDatetime

from scs_core.gas.a4.a4_calib import A4Calib
from scs_core.gas.a4.a4_datum import A4Datum


# TODO: return to this modelling technique using calib-age, when we have a wider range of calib-age training data?
# --------------------------------------------------------------------------------------------------------------------

class A4Calibrator(object):
    """
    classdocs
    """

    @classmethod
    def construct(cls, calib: A4Calib, era_start: LocalizedDatetime, era_end: LocalizedDatetime):
        era_length = era_end - era_start
        era_midpoint = era_end - (era_length / 2)

        return cls(calib, era_midpoint.timestamp())


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, calib: A4Calib, era_midpoint):
        """
        Constructor
        """
        self.__calib = calib                                                # A4Calib

        self.__we_elc_v = calib.we_elc_mv / 1000.0
        self.__ae_elc_v = calib.ae_elc_mv / 1000.0

        self.__we_sens_v = calib.we_sens_mv / 1000.0

        self.__we_no2_x_sens_v = None if calib.we_no2_x_sens_mv is None else calib.we_no2_x_sens_mv / 1000.0

        self.__era_midpoint = int(era_midpoint)


    # ----------------------------------------------------------------------------------------------------------------

    def train(self, datetime: LocalizedDatetime, datum, no2_cnc=None):
        timestamp = int(datetime.timestamp())
        cursor = timestamp - self.__era_midpoint

        return self.calibrate(datum, no2_cnc=no2_cnc, cursor=cursor)


    def calibrate(self, datum, no2_cnc=None, cursor=0):
        # zero offset...
        we_v_zero_cal = datum.we_v - self.__we_elc_v
        ae_v_zero_cal = datum.ae_v - self.__ae_elc_v

        # remove noise...
        v_zero_cal = we_v_zero_cal - ae_v_zero_cal

        # gain...
        v_cal = v_zero_cal / self.__we_sens_v

        # cross sensitivity...
        v_x_cal = None if no2_cnc is None else no2_cnc * self.__we_no2_x_sens_v

        return A4CalibratedDatum(datum.we_v, datum.ae_v, datum.we_c, datum.cnc,
                                 we_v_zero_cal, ae_v_zero_cal, v_cal, v_x_cal, cursor)


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "A4Calibrator:{calib:%s, era_midpoint:%s}" % (self.__calib, self.__era_midpoint)


# --------------------------------------------------------------------------------------------------------------------

class A4CalibratedDatum(A4Datum):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, we_v, ae_v, we_c, cnc, we_v_zero_cal, ae_v_zero_cal, v_cal, v_x_cal, cursor):
        """
        Constructor
        """
        super().__init__(we_v, ae_v, we_c, cnc)

        self.__we_v_zero_cal = Datum.float(we_v_zero_cal, 6)        # zero-offset-corrected WE voltage
        self.__ae_v_zero_cal = Datum.float(ae_v_zero_cal, 6)        # zero-offset-corrected AE voltage

        self.__v_cal = Datum.float(v_cal, 6)                        # calibrated voltage
        self.__v_x_cal = Datum.float(v_x_cal, 9)                    # calibrated cross-sensitivity voltage

        self.__cursor = int(cursor)                                 # timestamp relative to era midpoint


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['weV'] = self.we_v
        jdict['aeV'] = self.ae_v

        jdict['weC'] = self.we_c                                # may be None
        jdict['cnc'] = self.cnc                                 # may be None

        jdict['weVz'] = self.we_v_zero_cal
        jdict['aeVz'] = self.ae_v_zero_cal

        jdict['vCal'] = self.v_cal

        if self.v_x_cal is not None:
            jdict['vXCal'] = self.v_x_cal

        jdict['cursor'] = self.cursor

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def we_v_zero_cal(self):
        return self.__we_v_zero_cal


    @property
    def ae_v_zero_cal(self):
        return self.__ae_v_zero_cal


    @property
    def v_cal(self):
        return self.__v_cal


    @property
    def v_x_cal(self):
        return self.__v_x_cal


    @property
    def cursor(self):
        return self.__cursor


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "A4CalibratedDatum(vC):{we_v:%s, ae_v:%s, we_c:%s, cnc:%s, " \
               "we_v_zero_cal:%s, ae_v_zero_cal:%s, v_cal:%s, v_x_cal:%s, cursor:%s}" % \
               (self.we_v, self.ae_v, self.we_c, self.cnc,
                self.we_v_zero_cal, self.ae_v_zero_cal, self.v_cal, self.v_x_cal, self.cursor)



"""
Created on 5 Apr 2022

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example document (credentials):
{"username": "scs-opc-1", "password": "########"}

example document (identity):
{"username": "scs-bgx-401", "password": "########", "invoice": "INV-000123",
"created": "2023-06-23T10:32:52+01:00", "last-updated": "2023-06-23T10:32:52+01:00"}
"""

import re

from collections import OrderedDict

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONable

from scs_core.sys.logging import Logging
from scs_core.sys.shared_secret import SharedSecret
from scs_core.sys.system_id import SystemID


# --------------------------------------------------------------------------------------------------------------------

class CognitoDeviceCredentials(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def load_credentials_for_device(cls, host, strict=True):
        logger = Logging.getLogger()

        # SystemID...
        system_id = SystemID.load(host)

        if not system_id:
            if strict:
                logger.error("SystemID not available.")
                exit(1)
            else:
                return None

        # SharedSecret...
        shared_secret = SharedSecret.load(host)

        if not shared_secret:
            if strict:
                logger.error("SharedSecret not available.")
                exit(1)
            else:
                return None

        return cls(system_id.message_tag(), shared_secret.key)


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def is_valid_tag(cls, tag):
        try:
            if len(tag) > 255:
                return False

            return bool(re.fullmatch(r'[a-z]+\d*-[a-z]+\d*-\d+', tag))

        except TypeError:
            return False



    @classmethod
    def is_valid_password(cls, password):
        if not isinstance(password, str):
            return False

        return re.search(r'^\S{16,255}$', password) is not None


    @classmethod
    def multiple_tags(cls, prototype_tag, number):
        if not cls.is_valid_tag(prototype_tag):
            raise ValueError(prototype_tag)

        pieces = prototype_tag.split('-')

        vendor_id = pieces[0]
        model_id = pieces[1]
        system_serial = int(pieces[2])

        for system_serial_number in range(system_serial, system_serial + number):
            yield '-'.join((vendor_id, model_id, str(system_serial_number)))


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, tag, password):
        """
        Constructor
        """
        self._tag = tag                                         # PK: string
        self.__password = password                              # string


    def __lt__(self, other):
        return self.tag < other.tag


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['username'] = self.tag
        jdict['password'] = self.password

        return jdict

    # ----------------------------------------------------------------------------------------------------------------

    @property
    def username(self):
        return self.tag


    @property
    def tag(self):
        return self._tag


    @property
    def password(self):
        return self.__password


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "CognitoDeviceCredentials:{tag:%s, password:%s}" % \
               (self.tag, self.password)


# --------------------------------------------------------------------------------------------------------------------

class CognitoDeviceIdentity(CognitoDeviceCredentials):
    """
    classdocs
    """

    @classmethod
    def is_valid_invoice_number(cls, invoice_number):               # no longer in use - Larissa email 2024-02-06
        try:
            match = re.match(r'INV-\d{4,}', invoice_number)
            return match is not None

        except TypeError:
            return False


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_res(cls, res):
        if not res:
            return None

        tag = res.get('Username')
        password = res.get('password')

        invoice_number = res.get('invoice')

        created = round(LocalizedDatetime.construct_from_aws(str(res.get('UserCreateDate'))), 3).utc()
        last_updated = round(LocalizedDatetime.construct_from_aws(str(res.get('UserLastModifiedDate'))), 3).utc()

        return cls(tag, password=password, invoice_number=invoice_number, created=created, last_updated=last_updated)


    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return cls(None) if skeleton else None

        tag = jdict.get('username')
        password = jdict.get('password')

        invoice_number = jdict.get('invoice')

        created = LocalizedDatetime.construct_from_iso8601(jdict.get('created'))
        last_updated = LocalizedDatetime.construct_from_iso8601(jdict.get('last-updated'))

        return cls(tag, password=password, invoice_number=invoice_number, created=created, last_updated=last_updated)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, tag, password=None, invoice_number=None, created=None, last_updated=None):
        """
        Constructor
        """
        super().__init__(tag, password)

        self._invoice_number = invoice_number                   # string

        self._created = created                                 # LocalisedDatetime
        self._last_updated = last_updated                       # LocalizedDatetime


    def __eq__(self, other):
        try:
            return self.tag == other.tag

        except (TypeError, AttributeError):
            return False


    def __lt__(self, other):
        return self.tag < other.tag


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def index(self):
        return self.username


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['username'] = self.tag

        if self.password is not None:
            jdict['password'] = self.password

        jdict['invoice'] = self.invoice_number

        if self.created is not None:
            jdict['created'] = self.created.as_iso8601()

        if self.last_updated is not None:
            jdict['last-updated'] = self.last_updated.as_iso8601()

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def invoice_number(self):
        return self._invoice_number


    @property
    def created(self):
        return self._created


    @property
    def last_updated(self):
        return self._last_updated


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "CognitoDeviceIdentity:{tag:%s, password:%s, invoice_number:%s, created:%s, last_updated:%s}" % \
               (self.tag, self.password, self.invoice_number, self.created, self.last_updated)



"""
Created on 9 Jul 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://docs.python.org/3/library/multiprocessing.html#sharing-state-between-processes
https://eli.thegreenplace.net/2012/01/04/shared-counter-with-pythons-multiprocessing/
"""

from abc import ABC, abstractmethod

from multiprocessing import Process, Lock


# --------------------------------------------------------------------------------------------------------------------

class SynchronisedProcess(ABC):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, value=None):
        """
        Constructor
        """
        self.__proc = None
        self.__lock = Lock()
        self.__value = value


    # ----------------------------------------------------------------------------------------------------------------
    # process...

    def start(self):
        try:
            self.__proc = Process(target=self.run)
            self.__proc.start()

        except (ConnectionError, KeyboardInterrupt, SystemExit):
            pass


    def stop(self):
        try:
            if self.__proc is None:
                return

            self.__proc.terminate()

        except (ConnectionError, KeyboardInterrupt, SystemExit):
            pass


    def join(self):
        if self.__proc:
            self.__proc.join()


    @abstractmethod
    def run(self):
        pass


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def _proc(self):
        return self.__proc


    @property
    def _lock(self):
        return self.__lock


    @property
    def _value(self):
        return self.__value



"""
Created on 2 May 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

DEVICE    TYPE      STATE      CONNECTION
eth0      ethernet  connected  Ethernet eth0
cdc-wdm0  gsm       connected  giffgaff
sit0      iptunnel  unmanaged  --
lo        loopback  unmanaged  --

example JSON:
{"eth0": {"kind": "ethernet", "state": "connected", "connection": "Ethernet eth0"},
"cdc-wdm0": {"kind": "gsm", "state": "connected", "connection": "giffgaff"}}
"""

import re

from collections import OrderedDict

from scs_core.data.json import JSONable
from scs_core.data.str import Str


# --------------------------------------------------------------------------------------------------------------------

class Network(JSONable):
    """
    classdocs
    """

    __EXCLUDED_STATES = ('state', 'unmanaged')

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, device=None):
        if not jdict:
            return None

        kind = jdict.get('kind')
        state = jdict.get('state')
        connection = jdict.get('conn')

        return cls(device, kind, state, connection)


    @classmethod
    def construct_from_nmcli(cls, line):
        match = re.match(r'(\S+)\s+(\S+)\s+(\S+)\s+(\S.+)', line)

        if not match:
            return None

        state = match.groups()[2]

        if state.lower() in cls.__EXCLUDED_STATES:
            return None

        device = match.groups()[0]
        kind = match.groups()[1]

        reported_connection = match.groups()[3].strip()
        connection = None if reported_connection == '--' else reported_connection

        return cls(device, kind, state, connection)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, device, kind, state, connection):
        """
        Constructor
        """
        self.__device = device                      # string
        self.__kind = kind                          # string
        self.__state = state                        # string
        self.__connection = connection              # string


    def __eq__(self, other):
        try:
            return self.device == other.device and self.kind == other.kind and \
                   self.state == other.state and self.connection == other.connection

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def device(self):
        return self.__device


    @property
    def kind(self):
        return self.__kind


    @property
    def state(self):
        return self.__state


    @property
    def connection(self):
        return self.__connection


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['kind'] = self.kind
        jdict['state'] = self.state

        if self.connection:
            jdict['conn'] = self.connection

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Network:{device:%s, kind:%s, state:%s, connection:%s}" %  \
               (self.device, self.kind, self.state, self.connection)


# --------------------------------------------------------------------------------------------------------------------

class Networks(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        networks = OrderedDict()

        for device, network_jdict in jdict.items():
            networks[device] = Network.construct_from_jdict(network_jdict, device=device)

        return cls(networks)


    @classmethod
    def construct_from_nmcli(cls, lines):
        networks = OrderedDict()

        for line in lines:
            network = Network.construct_from_nmcli(line)

            if network:
                networks[network.device] = network

        return cls(networks)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, networks):
        """
        Constructor
        """
        self.__networks = networks                  # OrderedDict of device: Network


    def __len__(self):
        return len(self.networks)


    def __eq__(self, other):
        try:
            if len(self) != len(other):
                return False

            for device in self.__networks.keys():
                if self.__networks[device] != other.__networks[device]:
                    return False

            return True

        except (KeyError, TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def networks(self):
        return self.__networks


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        return self.networks


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Networks:{networks:%s}" %  Str.collection(self.networks)



"""
Created on 8 Mar 2021

@author: Jade Page (jade.page@southcoastscience.com)

A single manager for all required dynamoDB related functions.

https://stackoverflow.com/questions/36780856/complete-scan-of-dynamodb-with-boto3
https://github.com/boto/botocore/issues/1688
https://stackoverflow.com/questions/46616282/dynamodb-query-filterexpression-multiple-condition-chaining-python
"""

from boto3.dynamodb.conditions import Key, Attr, And
from functools import reduce

from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------


class DynamoManager(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, dynamo_client, dynamo_resource):
        self.__dynamo_client = dynamo_client
        self.__dynamo_resource = dynamo_resource


    # ----------------------------------------------------------------------------------------------------------------
    # SINGLE ITEMS
    # ----------------------------------------------------------------------------------------------------------------

    def get(self, table_name, pk, pk_val, sk=None, sk_val=None):
        # get specific item
        table = self.__dynamo_resource.Table(table_name)
        if sk and sk_val:
            key = {pk: pk_val, sk: sk_val}
        elif pk and pk_val:
            key = {pk: pk_val}
        else:
            return None

        response = table.get_item(Key=key)
        return response['Item'] if 'Item' in response else None


    def items(self, table_name, pk, pk_val, sk=None, sk_val=None):
        table = self.__dynamo_resource.Table(table_name)
        if sk is None and sk_val is None:
            response = table.query(
                KeyConditionExpression=Key(pk).eq(pk_val)
            )
        else:
            response = table.query(
                KeyConditionExpression=Key(pk).eq(pk_val) & Key(sk).eq(sk_val)
            )
        return response['Items'] if 'Items' in response else None


    def add(self, table_name, item):
        table = self.__dynamo_resource.Table(table_name)

        response = table.put_item(
            Item=item
        )

        return response


    def add_simple(self, table_name, pk, pk_val, sk=None, sk_val=None):
        table = self.__dynamo_resource.Table(table_name)
        if sk and sk_val:
            item = {pk: pk_val, sk: sk_val}
        elif pk and pk_val and not sk and not sk_val:
            item = {pk: pk_val}
        else:
            return None

        response = table.put_item(
            Item=item
        )

        return response


    def delete(self, table_name, pk, pk_val, sk=None, sk_val=None):
        table = self.__dynamo_resource.Table(table_name)
        if pk and pk_val and sk and sk_val:
            item = {pk: pk_val, sk: sk_val}
        elif pk and pk_val and not sk and not sk_val:
            item = {pk: pk_val}
        else:
            return None

        response = table.delete_item(
            Key=item
        )

        return response


    def delete_item(self, table_name, item):
        table = self.__dynamo_resource.Table(table_name)

        response = table.delete_item(
            Key=item
        )

        return response


    def update_item(self, table_name, pk, update_expression, eav):
        table = self.__dynamo_resource.Table(table_name)
        response = table.update_item(
            Key=pk,
            UpdateExpression=update_expression,
            ExpressionAttributeValues=eav,
            ReturnValues="UPDATED_NEW"
        )

        return response


    def filter_on_index(self, table_name, index_name, sk, filter_item):
        table = self.__dynamo_resource.Table(table_name)

        response = table.query(KeyConditionExpression=Key(sk).eq(filter_item),
                               IndexName=index_name)
        to_return = []
        to_return.extend(response['Items'])

        return to_return


    def multi_scan(self, table_name, key_value_pairs, limited):
        kwargs = {}

        if not table_name:
            return kwargs

        table = self.__dynamo_resource.Table(table_name)

        to_return = []
        q = table.scan(FilterExpression=reduce(And, ([Key(k).eq(v) for k, v in key_value_pairs])))
        to_return.extend(q['Items'])
        if not limited:
            while 'LastEvaluatedKey' in q:
                esk = q['LastEvaluatedKey']
                q = table.scan(FilterExpression=reduce(And, ([Key(k).eq(v) for k, v in key_value_pairs.items()])),
                               ExclusiveStartKey=esk)
                to_return.extend(q['Items'])
        return to_return


    # ----------------------------------------------------------------------------------------------------------------
    # MAIN FUNCTIONALITY
    # ----------------------------------------------------------------------------------------------------------------

    @staticmethod
    def build_query(table_name, pk=None, pk_val=None, sk=None, sk_val=None, keys_only=False, exact=False,
                    limit=None, fosk=False, ltsk=False):
        logger = Logging.getLogger()

        query = {}

        if not table_name:
            logger.debug('build_query: no table name')
            return query

        if ltsk:
            # Filter on less than second key - for expiring data
            logger.debug('build_query: type 1')
            query['KeyConditionExpression'] = Key(pk).eq(pk_val) & Key(sk).lt(sk_val)

        elif pk_val is None and sk is None and sk_val is None:
            # retrieve all - retrieve all batched - retrieve all keys only
            logger.debug('build_query: type 2')

        elif pk and pk_val and sk is None and sk_val is None and not fosk:
            # retrieve filtered on pk - retrieve filtered on pk keys only - retrieve filtered on pk batched
            logger.debug('build_query: type 3')
            if exact:
                query['KeyConditionExpression'] = (Key(pk).eq(pk_val))
            else:
                query['FilterExpression'] = (Attr(pk).contains(pk_val))

        elif pk and pk_val and sk and sk_val and not fosk:
            # retrieve double filtered - retrieve double filtered keys only - retrieve double filtered batched
            # the same as a get but can also be used with batch, exact and reserved keys
            logger.debug('build_query: type 4')
            if exact:
                query['KeyConditionExpression'] = Key(pk).eq(pk_val) & Key(sk).eq(sk_val)
            else:
                query['FilterExpression'] = Attr(pk).contains(pk_val) & Attr(sk).contains(sk_val)

        elif pk and sk and sk_val and fosk:
            # Filter on second key - filter on second key keys only - filter on second key batched
            logger.debug('build_query: type 5')
            if exact:
                query['FilterExpression'] = Key(sk).eq(sk_val)
            else:
                query['FilterExpression'] = Attr(sk).contains(sk_val)

        else:
            logger.debug('build_query: unrecognised type')

        if limit:
            query['Limit'] = limit

        if keys_only:
            query['ProjectionExpression'] = '#pk'
            query['ExpressionAttributeNames'] = {'#pk': pk}

        return query


    def do_query(self, table_name, query, limited=False):
        exact_match = False
        if "KeyConditionExpression" in query:
            exact_match = True
        table = self.__dynamo_resource.Table(table_name)
        to_return = []
        if exact_match:
            q = table.query(**query)
        else:
            q = table.scan(**query)
        to_return.extend(q['Items'])

        if not limited:
            while 'LastEvaluatedKey' in q:
                query['ExclusiveStartKey'] = q['LastEvaluatedKey']
                if exact_match:
                    q = table.query(**query)
                else:
                    q = table.scan(**query)
                to_return.extend(q['Items'])

        return to_return


    def do_query_batched(self, table_name, query, exclusive_start_key=None):
        exact_match = False
        if "KeyConditionExpression" in query:
            exact_match = True
        last_evaluated_key = None
        table = self.__dynamo_resource.Table(table_name)
        to_return = []
        if exclusive_start_key:
            query['ExclusiveStartKey'] = exclusive_start_key
        if exact_match:
            q = table.query(**query)
        else:
            q = table.scan(**query)
        to_return.extend(q['Items'])
        if 'LastEvaluatedKey' in q:
            last_evaluated_key = q['LastEvaluatedKey']

        return to_return, last_evaluated_key


    # ----------------------------------------------------------------------------------------------------------------
    # UTILITIES
    # ----------------------------------------------------------------------------------------------------------------

    def describe_table(self, table_name):
        response = self.__dynamo_client.describe_table(
            TableName=table_name
        )

        return response


    def return_key_schema(self, table_name):
        table = self.__dynamo_resource.Table(table_name)
        return table.key_schema


    # ----------------------------------------------------------------------------------------------------------------
    # BATCH FUNCTIONS
    # ----------------------------------------------------------------------------------------------------------------

    def batch_delete(self, table_name, items):
        table = self.__dynamo_resource.Table(table_name)
        with table.batch_writer() as batch:
            for item in items:
                batch.delete_item(item)


    def batch_write(self, table_name, items):
        table = self.__dynamo_resource.Table(table_name)
        with table.batch_writer() as batch:
            for item in items:
                batch.put_item(Item=item)



"""
Created on 7 Oct 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example document:
{"location-path": "southcoastscience-dev/test/loc/1", "device-path": "southcoastscience-dev/test/device"}
"""
import string

from collections import OrderedDict

from scs_core.data.json import PersistentJSONable


# --------------------------------------------------------------------------------------------------------------------

class Project(PersistentJSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    __FILENAME = "aws_project.json"

    @classmethod
    def persistence_location(cls):
        return cls.aws_dir(), cls.__FILENAME


    # ----------------------------------------------------------------------------------------------------------------

    ENVIRONMENT_CHANNELS =  ('C', 'G', 'P')
    SYSTEM_CHANNELS =  ('S', 'X')

    CHANNELS = ENVIRONMENT_CHANNELS + SYSTEM_CHANNELS

    @classmethod
    def is_valid_channel(cls, channel):
        return channel in cls.CHANNELS


    # ----------------------------------------------------------------------------------------------------------------

    PERMITTED_CHARS = set(string.ascii_letters + string.digits + ' ' + '-' + '_')       # ' ' will be substituted

    @classmethod
    def is_valid_path_node(cls, chars):
        if not chars:
            return False

        return set(chars) <= cls.PERMITTED_CHARS


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, organisation, group, location):
        std_organisation = organisation.replace(' ', '-').lower()
        std_group = group.replace(' ', '-').lower()
        std_location = str(location).replace(' ', '-')

        location_path = '/'.join((std_organisation, std_group, 'loc', std_location))
        device_path = '/'.join((std_organisation, std_group, 'device'))

        return Project(location_path, device_path)


    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        location_path = jdict.get('location-path')
        device_path = jdict.get('device-path')

        return Project(location_path, device_path)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, location_path, device_path):
        """
        Constructor
        """
        super().__init__()

        self.__location_path = location_path            # string
        self.__device_path = device_path                # string


    def __eq__(self, other):
        try:
            return self.location_path == other.location_path and self.device_path == other.device_path

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['location-path'] = self.__location_path
        jdict['device-path'] = self.__device_path

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def environment_paths(self, system_id):
        return tuple(self.channel_path(channel, system_id) for channel in self.ENVIRONMENT_CHANNELS)


    def channel_path(self, channel, system_id):
        if channel == 'C':
            return self.subject_path('climate', system_id)

        if channel == 'G':
            return self.subject_path('gases', system_id)

        if channel == 'P':
            return self.subject_path('particulates', system_id)

        if channel == 'S':
            return self.subject_path('status', system_id)

        if channel == 'X':
            return self.subject_path('control', system_id)

        raise ValueError(channel)


    def subject_path(self, subject, system_id):
        if subject == 'climate' or subject == 'gases' or subject == 'particulates':
            return '/'.join((self.__location_path, subject))

        if subject == 'status' or subject == 'control':
            return '/'.join((self.__device_path, system_id.topic_label(), subject))

        return None


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def location_path(self):
        return self.__location_path


    @property
    def device_path(self):
        return self.__device_path


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Project:{location_path:%s, device_path:%s}" % (self.location_path, self.device_path)



"""
Created on 14 Jan 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://stackoverflow.com/questions/51700960/runtimeerror-generator-raised-stopiteration-every-time-i-try-to-run-app
"""

import sys
import time

from collections import OrderedDict
from multiprocessing import Manager

from requests.exceptions import ConnectionError

from scs_core.aws.manager.byline.byline_finder import DeviceBylineFinder
from scs_core.aws.security.cognito_login_manager import CognitoLoginManager

from scs_core.csv.csv_log_cursor_queue import CSVLogCursorQueue, CSVLogCursor
from scs_core.csv.csv_reader import CSVReader

from scs_core.sync.synchronised_process import SynchronisedProcess

from scs_core.sys.logging import Logging
from scs_core.sys.tail import Tail


# --------------------------------------------------------------------------------------------------------------------

class CSVLogReader(SynchronisedProcess):
    """
    classdocs
    """

    __IDLE_TIME =       4.0                 # seconds

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def __read_rows(cls, reader):
        try:
            for datum in reader.rows():
                print(datum)
                sys.stdout.flush()

        except (ConnectionError, KeyboardInterrupt, SystemExit):
            pass


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, queue_builder, nullify=False):
        """
        Constructor
        """
        self.__logger = Logging.getLogger()
        self.__logging_specification = Logging.specification()

        manager = Manager()

        SynchronisedProcess.__init__(self, value=manager.list())

        queue = CSVLogCursorQueue()

        with self._lock:
            queue.as_list(self._value)

        self.__queue_builder = queue_builder                                    # CSVLogQueueBuilder
        self.__nullify = bool(nullify)                                          # bool


    # ----------------------------------------------------------------------------------------------------------------

    def run(self, halt_on_empty_queue=False):
        Logging.replicate(self.__logging_specification)
        self.__logger = Logging.getLogger()

        try:
            # build queue...
            timeline_start, cursors = self.__queue_builder.find_cursors()       # waits indefinitely for network

            self.__logger.info("timeline_start: %s" % timeline_start)

            with self._lock:
                queue = CSVLogCursorQueue.construct_from_jdict(OrderedDict(self._value))
                for cursor in cursors:
                    queue.include(cursor)
                queue.as_list(self._value)

            # drain queue...
            while True:
                # find oldest...
                with self._lock:
                    queue = CSVLogCursorQueue.construct_from_jdict(OrderedDict(self._value))
                    cursor = queue.next()
                    queue.as_list(self._value)

                if cursor is None:
                    if halt_on_empty_queue:
                        return
                    else:
                        time.sleep(self.__IDLE_TIME)
                        continue

                self.__logger.info("reading: %s" % cursor)

                # process...
                read_count = self.__tail(cursor) if cursor.is_live else self.__read(cursor)

                self.__logger.info("closing: %s: read: %s" % (cursor, read_count))

                # remove...
                with self._lock:
                    queue = CSVLogCursorQueue.construct_from_jdict(OrderedDict(self._value))
                    queue.remove(cursor.file_path)
                    queue.as_list(self._value)

                # self.__logger.info("*** removed cursor: %s" % cursor)

                if not cursor.is_live:
                    continue

                # try tailed file again...
                with self._lock:
                    tailed = CSVLogCursor(cursor.file_path, read_count, False)
                    queue = CSVLogCursorQueue.construct_from_jdict(OrderedDict(self._value))
                    queue.include(tailed)
                    queue.as_list(self._value)

                self.__logger.info("added tailed cursor: %s" % tailed)

        except FileNotFoundError as ex:
            self.__logger.error(repr(ex))

        except (ConnectionError, EOFError, KeyboardInterrupt, SystemExit):
            pass


    # ----------------------------------------------------------------------------------------------------------------
    # run methods...

    def __read(self, cursor: CSVLogCursor):
        reader = CSVReader.construct_for_file(cursor.file_path, nullify=self.__nullify, start_row=cursor.row)
        try:
            self.__read_rows(reader)

        finally:
            reader.close()

            return reader.read_count


    def __tail(self, cursor: CSVLogCursor):
        tail = Tail.construct(cursor.file_path)
        tail.open()

        reader = CSVReader(tail, nullify=self.__nullify, start_row=cursor.row)
        try:
            self.__read_rows(reader)

        except RuntimeError:            # Python 3.7 response to StopIteration
            pass                        # file is closed - but were all the rows read?!?

        except TimeoutError:
            self.__logger.error("TimeoutError: %s: read: %s" % (cursor, reader.read_count))

        finally:
            reader.close()
            tail.close()

            return reader.read_count


    # ----------------------------------------------------------------------------------------------------------------
    # setters for client process...

    def include(self, file_path):
        if file_path is None:
            return

        cursor = CSVLogCursor(file_path, 0, True)

        with self._lock:
            queue = CSVLogCursorQueue.construct_from_jdict(OrderedDict(self._value))
            queue.include(cursor)
            queue.as_list(self._value)


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        with self._lock:
            queue = CSVLogCursorQueue.construct_from_jdict(OrderedDict(self._value))

        return "CSVLogReader:{queue_builder:%s, queue:%s, nullify:%s}" % \
               (self.__queue_builder, queue, self.__nullify)


# --------------------------------------------------------------------------------------------------------------------

class CSVLogQueueBuilder(object):
    """
    classdocs
    """

    __BYLINE_WAIT_TIME = 30.0           # seconds

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, conf, credentials, message_tag, topic_name, topic_path):
        """
        Constructor
        """
        self.__conf = conf                                      # CSVLoggerConf
        self.__credentials = credentials                        # CognitoDeviceCredentials
        self.__message_tag = message_tag                        # string
        self.__topic_name = topic_name                          # string
        self.__topic_path = topic_path                          # string

        self.__logger = Logging.getLogger()


    # ----------------------------------------------------------------------------------------------------------------

    def find_cursors(self):
        gatekeeper = CognitoLoginManager()
        finder = DeviceBylineFinder()

        # timeline_start...
        while True:
            try:
                auth = gatekeeper.device_login(self.__credentials)

                if not auth.is_ok():
                    self.__logger.error(auth.authentication_status.description)

                byline = finder.find_byline_for_topic(auth.id_token, self.__topic_path, include_messages=False)
                break

            except ConnectionError as ex:
                self.__logger.info(ex)
                time.sleep(self.__BYLINE_WAIT_TIME)

        rec = None if byline is None else byline.rec
        byline_start = None if rec is None else rec.utc_datetime

        timeline_start = self.__conf.utc_retrospection_start(byline_start)

        # CSVLog...
        read_log = self.__conf.csv_log(self.__topic_name, tag=self.__message_tag, timeline_start=timeline_start)

        return timeline_start, CSVLogCursorQueue.find_cursors_for_log(read_log, 'rec')  # may raise FileNotFoundError


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self):
        return "CSVLogQueueBuilder:{conf:%s, credentials:%s, message_tag:%s, topic_name:%s, topic_path:%s}" % \
               (self.__conf, self.__credentials, self.__message_tag, self.__topic_name, self.__topic_path)



"""
Created on 19 Sep 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

A catalogue entry for a machine learning model, for a specific gas

https://github.com/south-coast-science/scs_sagemaker_automl/blob/a1db0b830909b20fc3a7cea8b895a8bdb8d5b033/src/scs_sagemaker/project.py#L277

IMPORTANT: Assuming that the miniumum gas concentration in the data set is close to zero, the minimum vCal value
of the primary term indicates the minimum vCal value for every equivalent sensor - the vCal baseline on the target
devices should be set to make this so.

document example:
{"is-error-model": true,
"data-set": "ref-scs-bgx-642-gases-21Q3-slp16-clean-vcal-err",
"period": {"start": "2021-11-01T00:02:00Z", "end": "2021-12-28T06:57:00Z"},
"primaries": {"SO2.vCal": {"path": "src.scs.env.gases.val.SO2.vCal", "min": -355.462, "avg": 18.488, "max": 4256.394}},
"secondaries": {"tmp.brd": {"path": "src.scs.stat.val.tmp.brd", "min": 21.4, "avg": 31.994, "max": 48.7},
"hmd.cur": {"path": "src.scs.env.meteo.val.hmd.cur", "min": 11.0, "avg": 51.815, "max": 84.1},
"hmd.slope": {"path": "src.scs.env.meteo.val.hmd.slope", "min": -0.022, "avg": 0.0, "max": 0.031},
"tmp.cur": {"path": "src.scs.env.meteo.val.tmp.cur", "min": 16.3, "avg": 26.568, "max": 42.8},
"tmp.slope": {"path": "src.scs.env.meteo.val.tmp.slope", "min": -0.009, "avg": -0.0, "max": 0.003}},
"reference": {"path": "src.ref.gas.SO2", "min": 1.0, "avg": 18.5, "max": 3211.0},
"output": {"path": "exg.SO2.vE.Urban.21Q3", "min": -63.2, "avg": 18.6, "max": 4228.2},
"performance": {"count": 78052, "slope": 0.953, "intercept": 0.999, "r2": 0.863, "p": 0.0, "std-err": 0.001}}
"""

import os

from collections import OrderedDict

from scs_core.data.json import JSONCatalogueEntry, JSONify
from scs_core.data.lin_regress import LinRegress
from scs_core.data.path_dict import PathDict
from scs_core.data.str import Str

from scs_core.model.catalogue.term import Term, PrimaryTerm, SecondaryTerm
from scs_core.model.catalogue.training_period import TrainingPeriod

from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

class ModelCompendium(JSONCatalogueEntry):
    """
    classdocs
    """

    __CATALOGUE_NAME = 'compendia'

    @classmethod
    def catalogue_location(cls):
        return os.path.join(os.path.dirname(os.path.realpath(__file__)), cls.__CATALOGUE_NAME)


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def __term_path(cls, path):
        return '.'.join(path.split('.')[-2:])


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, is_error_model, data_set, recs, primary_values, secondary_values, output_path, outputs,
                  reference_path, references):
        period = TrainingPeriod.construct(recs)

        primaries = {cls.__term_path(path): PrimaryTerm.construct(path, values, prec=3)
                     for path, values in primary_values.items()}

        secondaries = {cls.__term_path(path): SecondaryTerm.construct(path, values, prec=3)
                       for path, values in secondary_values.items()}

        reference = Term.construct(reference_path, references, prec=1)
        output = Term.construct(output_path, outputs, prec=1)

        performance = LinRegress.construct(references, outputs, prec=3)

        return cls(is_error_model, data_set, period, primaries, secondaries, reference, output, performance)


    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        is_error_model = jdict.get('is-error-model', False)
        data_set = jdict.get('data-set')
        period = TrainingPeriod.construct_from_jdict(jdict.get('period'))

        primaries = {path: PrimaryTerm.construct_from_jdict(term_jdict)
                     for path, term_jdict in jdict.get('primaries').items()}

        secondaries = {path: SecondaryTerm.construct_from_jdict(term_jdict)
                       for path, term_jdict in jdict.get('secondaries').items()}

        reference = Term.construct_from_jdict(jdict.get('reference'))
        output = Term.construct_from_jdict(jdict.get('output'))

        performance = LinRegress.construct_from_jdict(jdict.get('performance'))

        return cls(is_error_model, data_set, period, primaries, secondaries, reference, output, performance)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, is_error_model, data_set, period, primaries, secondaries, reference, output, performance):
        """
        Constructor
        """
        self.__is_error_model = bool(is_error_model)        # bool
        self.__data_set = data_set                          # string
        self.__period = period                              # TrainingPeriod

        self.__primaries = primaries                        # dict of path: PrimaryTerm
        self.__secondaries = secondaries                    # dict of path: SecondaryTerm
        self.__reference = reference                        # Term
        self.__output = output                              # Term

        self.__performance = performance                    # LinRegress

        self.__logger = Logging.getLogger()


    def __len__(self):
        return len(self.performance)


    # ----------------------------------------------------------------------------------------------------------------

    def is_in_bounds(self, datum: PathDict):
        for datum_path in datum.paths():
            term_path = self.__term_path(datum_path)

            if term_path in self.secondaries:
                if not self.secondaries[term_path].is_in_bounds(datum.node(datum_path)):
                    return False

        return True


    def preprocess(self, datum: PathDict, offset=0, include_orig=False):
        target = PathDict()

        for datum_path in datum.paths():
            term_path = self.__term_path(datum_path)
            node = datum.node(datum_path)

            if term_path in self.primaries:
                value, extr = self.primaries[term_path].preprocess(node, offset)

                if include_orig:
                    target.append(datum_path + 'Orig', node)

                target.append(datum_path, None if value is None else "%g" % value)
                target.append(datum_path + 'Extr', None if value is None else "%g" % extr)
                continue

            if term_path in self.secondaries:
                value = self.secondaries[term_path].preprocess(node)

                if include_orig:
                    target.append(datum_path + 'Orig', node)

                target.append(datum_path, None if value is None else "%g" % value)
                continue

            target.append(datum_path, node)

        self.__logger.debug("preprocess - datum: %s" % JSONify.dumps(datum))
        self.__logger.debug("preprocess - target: %s" % JSONify.dumps(target))

        return target


    def postprocess(self, primary_path, vcal_excess, model_output):
        # self.__logger.info("postprocess - primary_path: %s vcal_excess: %s model_output: %s" %
        #                    (primary_path, vcal_excess, model_output))

        corrected_exg = (model_output - self.performance.intercept) / self.performance.slope

        # TODO: handle vCal and xCal excesses differently

        if vcal_excess > 0:
            model_vcal_max = self.primary_term(primary_path).maximum
            corrected_exg = corrected_exg * ((model_vcal_max + vcal_excess) / model_vcal_max)

        return corrected_exg


    # ----------------------------------------------------------------------------------------------------------------

    def primary_paths(self):
        return [primary.path for primary in self.primaries.values()]


    def primary_term(self, path):
        try:
            return self.primaries[self.__term_path(path)]
        except KeyError:
            return None


    def secondary_paths(self):
        return [secondary.path for secondary in self.secondaries.values()]


    def secondary_term(self, path):
        try:
            return self.secondaries[self.__term_path(path)]
        except KeyError:
            return None


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def name(self):
        pieces = self.output.path.split('.')
        return '.'.join(pieces[1:])


    @property
    def species_name(self):
        pieces = self.output.path.split('.')
        return pieces[1]


    @property
    def model_name(self):
        pieces = self.output.path.split('.')
        return pieces[2]


    @property
    def device_name(self):
        pieces = self.output.path.split('.')
        return pieces[3]


    @property
    def period_name(self):
        pieces = self.output.path.split('.')
        return pieces[4]


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def is_error_model(self):
        return self.__is_error_model


    @property
    def data_set(self):
        return self.__data_set


    @property
    def period(self):
        return self.__period


    @property
    def primaries(self):
        return self.__primaries


    @property
    def secondaries(self):
        return self.__secondaries


    @property
    def reference(self):
        return self.__reference


    @property
    def output(self):
        return self.__output


    @property
    def performance(self):
        return self.__performance


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['is-error-model'] = self.is_error_model
        jdict['data-set'] = self.data_set
        jdict['period'] = self.period

        jdict['primaries'] = self.primaries
        jdict['secondaries'] = self.secondaries
        jdict['reference'] = self.reference
        jdict['output'] = self.output

        jdict['performance'] = self.performance

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        primaries = Str.collection(self.primaries)
        secondaries = Str.collection(self.secondaries)

        return "ModelCompendium:{name:%s, is_error_model:%s, data_set:%s, period:%s, primaries:%s, secondaries:%s, " \
               "reference:%s, output:%s, performance:%s}" %  \
               (self.name, self.is_error_model, self.data_set, self.period, primaries, secondaries,
                self.reference, self.output, self.performance)



"""
Created on 14 Nov 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example document:
{"rec": "2021-10-11T11:11:14Z", "tag": "scs-be2-3", "ver": 2.0, "src": "N3",
"val": {"per": 4.1, "pm1": 0.7, "pm2p5": 3.2, "pm10": 33.3,
"bin": [50, 24, 6, 2, 4, 2, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
"mtf1": 29, "mtf3": 32, "mtf5": 19, "mtf7": 54, "sfr": 3.73,
"sht": {"hmd": 43.6, "tmp": 25.6}},
"exg": {"src": "rn20", "val": {"pm1": 1.5, "pm2p5": 6.4, "pm10": 58.7}}}
"""

from collections import OrderedDict

from scs_core.data.str import Str

from scs_core.particulate.opc_datum import OPCDatum

from scs_core.sample.sample import Sample


# --------------------------------------------------------------------------------------------------------------------

class ParticulatesSample(Sample):
    """
    classdocs
    """

    VERSION = 1.0

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        # Sample...
        tag = jdict.get('tag')

        try:
            version = round(float(jdict.get('ver')), 1)
        except (TypeError, ValueError):
            version = cls.DEFAULT_VERSION

        val = jdict.get('val')
        exegeses = jdict.get('exg')

        # ParticulatesSample...
        val['rec'] = jdict.get('rec')
        val['src'] = jdict.get('src')

        opc_datum = OPCDatum.construct_from_jdict(val)

        return cls(tag, opc_datum, version=version, exegeses=exegeses)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, tag, opc_datum, version=None, exegeses=None):
        """
        Constructor
        """
        if version is None:
            version = self.VERSION

        super().__init__(tag, opc_datum.rec, version, src=opc_datum.source, exegeses=exegeses)

        self.__opc_datum = opc_datum                                # OPCDatum


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def has_invalid_value(cls):
        # TODO: implement has_invalid_value
        return False


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def values(self):
        jdict = OrderedDict()

        jdict['per'] = self.opc_datum.period

        jdict['pm1'] = self.opc_datum.pm1
        jdict['pm2p5'] = self.opc_datum.pm2p5
        jdict['pm10'] = self.opc_datum.pm10

        jdict['bin'] = self.opc_datum.bins

        jdict['mtf1'] = self.opc_datum.bin_1_mtof
        jdict['mtf3'] = self.opc_datum.bin_3_mtof
        jdict['mtf5'] = self.opc_datum.bin_5_mtof
        jdict['mtf7'] = self.opc_datum.bin_7_mtof

        if self.opc_datum.sfr is not None:
            jdict['sfr'] = self.opc_datum.sfr

        if self.opc_datum.sht is not None:
            jdict['sht'] = self.opc_datum.sht.as_json()

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def opc_datum(self):
        return self.__opc_datum


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        exegeses = Str.collection(self.exegeses)

        return "ParticulatesSample:{tag:%s, rec:%s, src:%s, exegeses:%s, opc_datum:%s}" % \
            (self.tag, self.rec, self.src, exegeses, self.opc_datum)



"""
Created on 24 Mar 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

Modem
-----
modem.generic.device-identifier                 : 3f07553c31ce11715037ac16c24ceddcfb6f7a0b
modem.generic.manufacturer                      : QUALCOMM INCORPORATED
modem.generic.model                             : QUECTEL Mobile Broadband Module
modem.generic.revision                          : EC21EFAR06A01M4G
...
modem.3gpp.imei                                 : 867962041294151

example JSON:
{"id": "3f07553c31ce11715037ac16c24ceddcfb6f7a0b", "imei": "867962041294151", "mfr": "QUALCOMM INCORPORATED",
"model": "QUECTEL Mobile Broadband Module", "rev": "EC21EFAR06A01M4G"}


ModemConnection
---------------
modem.generic.state                         : connected
modem.generic.state-failed-reason           : --
modem.generic.signal-quality.value          : 67
modem.generic.signal-quality.recent         : yes

example JSON:
{"state": "connected", "signal": {"quality": 67, "recent": true}}


SIM (Subscriber Identity Module)
--------------------------------
sim.dbus-path                               : /org/freedesktop/ModemManager1/SIM/0
sim.properties.imsi                         : 234104886708667
sim.properties.iccid                        : 8944110068256270054
sim.properties.operator-code                : 23410
sim.properties.operator-name                : giffgaff
sim.properties.emergency-numbers.length     : 2
sim.properties.emergency-numbers.value[1]   : 999
sim.properties.emergency-numbers.value[2]   : 00112

example JSON:
{"imsi": "123", "iccid": "456", "operator-code": "789 012", "operator-name": "giff gaff"}
"""

import re

from collections import OrderedDict

from scs_core.data.datum import Datum
from scs_core.data.json import JSONable

from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

class ModemList(object):
    """
    modem-list.value[1] : /org/freedesktop/ModemManager1/Modem/0
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_mmcli(cls, lines):
        modems = []

        for line in lines:
            match = re.match(r'modem-list.value\[\d+]\s+:\s+(\S+)', line)
            if match:
                modems.append(match.groups()[0])

        return cls(modems)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, modems):
        """
        Constructor
        """
        self.__modems = modems                              # array of string


    def __len__(self):
        return len(self.__modems)


    # ----------------------------------------------------------------------------------------------------------------

    def modem(self, index):
        return self.__modems[index]


    def number(self, index):
        pieces = self.modem(index).split('/')

        return pieces[-1]


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ModemList:{modems:%s}" % self.__modems


# --------------------------------------------------------------------------------------------------------------------

class Modem(JSONable):
    """
    modem.generic.device-identifier                 : 3f07553c31ce11715037ac16c24ceddcfb6f7a0b
    modem.generic.manufacturer                      : QUALCOMM INCORPORATED
    modem.generic.revision                          : EC21EFAR06A01M4G
    ...
    modem.3gpp.imei                                 : 867962041294151
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        id = jdict.get('id')
        imei = jdict.get('imei')
        mfr = jdict.get('mfr')
        rev = jdict.get('rev')

        return cls(id, imei, mfr, rev)


    @classmethod
    def construct_from_mmcli(cls, lines):
        id = None
        imei = None
        mfr = None
        rev = None

        for line in lines:
            match = re.match(r'modem\.generic\.device-identifier\s+:\s+(\S+)', line)
            if match:
                id = match.groups()[0]
                continue

            match = re.match(r'.*\.imei\s+:\s+(\d+)', line)
            if match:
                imei = match.groups()[0]
                continue

            match = re.match(r'modem\.generic\.manufacturer\s+:\s+(\S.*\S)', line)
            if match:
                mfr = match.groups()[0]
                continue

            match = re.match(r'modem\.generic\.revision\s+:\s+(\S+)', line)
            if match:
                rev = match.groups()[0]
                continue

        return cls(id, imei, mfr, rev)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, id, imei, mfr, rev):
        """
        Constructor
        """
        self.__id = id                              # string
        self.__imei = imei                          # string
        self.__mfr = mfr                            # string
        self.__rev = rev                            # string


    def __eq__(self, other):
        try:
            return self.id == other.id and self.imei == other.imei and self.mfr == other.mfr and \
                   self.rev == other.rev

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def id(self):
        return self.__id


    @property
    def imei(self):
        return self.__imei


    @property
    def mfr(self):
        return self.__mfr


    @property
    def rev(self):
        return self.__rev


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['id'] = self.id
        jdict['imei'] = self.imei
        jdict['mfr'] = self.mfr
        jdict['rev'] = self.rev

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Modem:{id:%s, imei:%s, mfr:%s, rev:%s}" %  \
               (self.id, self.imei, self.mfr, self.rev)


# --------------------------------------------------------------------------------------------------------------------

class ModemConnection(JSONable):
    """
    modem.generic.state                         : connected
    modem.generic.state-failed-reason           : --
    modem.generic.signal-quality.value          : 67
    modem.generic.signal-quality.recent         : yes
    """

    UNAVAILABLE_STATE = "unavailable"

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        state = jdict.get('state')
        failure = jdict.get('failure')
        signal = Signal.construct_from_jdict(jdict.get('signal'))

        return cls(state, failure, signal)


    @classmethod
    def construct_from_mmcli(cls, lines):
        first_line = None
        state = None
        failure = None
        quality = None
        recent = None

        for line in lines:
            if first_line is None:
                first_line = line

            match = re.match(r'modem\.generic\.state\s+:\s+([a-z]+)', line)
            if match:
                state = match.groups()[0]
                continue

            match = re.match(r'modem\.generic\.state-failed-reason\s+:\s+(\S.*\S)', line)
            if match:
                reported_failure = match.groups()[0]
                failure = None if reported_failure == '--' else reported_failure
                continue

            match = re.match(r'modem\.generic\.signal-quality\.value\s+:\s+(\d+)', line)
            if match:
                quality = match.groups()[0]
                continue

            match = re.match(r'modem\.generic\.signal-quality\.recent\s+:\s+([a-z]+)', line)
            if match:
                recent = match.groups()[0] == 'yes'
                continue

        if state is None:
            Logging.getLogger().error("construct_from_mmcli: could not get state from '%s'" % first_line)

        return cls(state, failure, Signal(quality, recent))


    @classmethod
    def null_datum(cls):
        return cls(cls.UNAVAILABLE_STATE, None, Signal.null_datum())


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, state, failure, signal):
        """
        Constructor
        """
        self.__state = state                            # string
        self.__failure = failure                        # string
        self.__signal = signal                          # Signal


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def state(self):
        return self.__state


    @property
    def failure(self):
        return self.__failure


    @property
    def signal(self):
        return self.__signal


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['state'] = self.state

        if self.failure is not None:
            jdict['failure'] = self.failure

        jdict['signal'] = self.signal

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ModemConnection:{state:%s, failure:%s, signal:%s}" %  (self.state, self.failure, self.signal)


# --------------------------------------------------------------------------------------------------------------------

class Signal(JSONable):
    """
    modem.generic.signal-quality.value          : 67
    modem.generic.signal-quality.recent         : yes
    """

    __SIGNIFICANT_QUALITY_DIFFERENCE = 10

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        quality = jdict.get('quality')
        recent = jdict.get('recent')

        return cls(quality, recent)


    @classmethod
    def null_datum(cls):
        return cls(None, None)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, quality, recent):
        """
        Constructor
        """
        self.__quality = Datum.int(quality)                 # int
        self.__recent = recent                              # bool


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def quality(self):
        return self.__quality


    @property
    def recent(self):
        return self.__recent


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['quality'] = self.quality
        jdict['recent'] = self.recent

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Signal:{quality:%s, recent:%s}" %  (self.quality, self.recent)


# --------------------------------------------------------------------------------------------------------------------

class SIMList(object):
    """
    modem.generic.sim    : /org/freedesktop/ModemManager1/SIM/0
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_mmcli(cls, lines):
        sims = []

        for line in lines:
            match = re.match(r'modem\.generic\.sim\s+:\s+(\S+)', line)
            if match:
                sims.append(match.groups()[0])

        return cls(sims)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, sims):
        """
        Constructor
        """
        self.__sims = sims                                  # array of string


    def __len__(self):
        return len(self.__sims)


    # ----------------------------------------------------------------------------------------------------------------

    def sim(self, index):
        return self.__sims[index]


    def number(self, index):
        pieces = self.sim(index).split('/')

        return pieces[-1]


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "SIMList:{sims:%s}" % self.__sims


# --------------------------------------------------------------------------------------------------------------------

class SIM(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        imsi = jdict.get('imsi')
        iccid = jdict.get('iccid')
        operator_code = jdict.get('operator-code')
        operator_name = jdict.get('operator-name')

        return cls(imsi, iccid, operator_code, operator_name)


    @classmethod
    def construct_from_mmcli(cls, lines):
        imsi = None
        iccid = None
        operator_code = None
        operator_name = None

        for line in lines:
            match = re.match(r'sim\.properties\.imsi\s+:\s+(\d+)', line)
            if match:
                imsi = match.groups()[0]
                continue

            match = re.match(r'sim\.properties\.iccid\s+:\s+(\d+)', line)
            if match:
                iccid = match.groups()[0]
                continue

            match = re.match(r'sim\.properties\.operator-code\s+:\s+(\d+)', line)
            if match:
                operator_code = match.groups()[0]
                continue

            match = re.match(r'sim\.properties\.operator-name\s+:\s+(\S.*)', line)
            if match:
                reported_name = match.groups()[0].strip()
                operator_name = None if reported_name == '--' else reported_name

        return cls(imsi, iccid, operator_code, operator_name)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, imsi, iccid, operator_code, operator_name):
        """
        Constructor
        """
        self.__imsi = imsi                                  # numeric string
        self.__iccid = iccid                                # numeric string
        self.__operator_code = operator_code                # string
        self.__operator_name = operator_name                # string


    def __eq__(self, other):
        try:
            return self.imsi == other.imsi and self.iccid == other.iccid and \
                   self.operator_code == other.operator_code and self.operator_name == other.operator_name

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def imsi(self):
        return self.__imsi


    @property
    def iccid(self):
        return self.__iccid


    @property
    def operator_code(self):
        return self.__operator_code


    @property
    def operator_name(self):
        return self.__operator_name


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['imsi'] = str(self.imsi)
        jdict['iccid'] = str(self.iccid)
        jdict['operator-code'] = self.operator_code
        jdict['operator-name'] = self.operator_name

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "SIM:{imsi:%s, iccid:%s, operator_code:%s, operator_name:%s}" % \
               (self.imsi, self.iccid, self.operator_code, self.operator_name)



"""
Created on 12 Apr 2018

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example FilesystemReport JSON:
{"path": "/srv/removable_data_storage", "is-available": true, "on-root": true, "used": 69}
"""

import os

from collections import OrderedDict

from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class Filesystem(object):

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def mkdir(cls, path):
        if not path or os.path.exists(path):
            return

        head, _ = os.path.split(path)

        if head and not os.path.exists(head):
            cls.mkdir(head)

        if not os.path.exists(path):                # handles case of trailing /
            try:
                os.mkdir(path)
            except FileExistsError:
                raise FileNotFoundError(path)       # covers a Python bug


    @classmethod
    def rmdir(cls, path):
        if not path or not os.path.exists(path):
            return False

        try:
            os.rmdir(path)
            return True

        except OSError:
            return False


    @classmethod
    def rm(cls, path):
        if not path or not os.path.exists(path):
            return False

        try:
            os.remove(path)
            return True

        except OSError:
            return False


    @classmethod
    def ls(cls, path):
        if not path or not os.path.exists(path):
            return None

        return tuple(File.construct(path, name) for name in sorted(os.listdir(path)))


    @classmethod
    def is_on_root_filesystem(cls, path):
        ds_rp = os.path.realpath(path)

        return os.stat(ds_rp).st_dev == os.stat('/').st_dev                     # may raise FileNotFoundError


    @classmethod
    def percentage_used(cls, path):
        ds_rp = os.path.realpath(path)
        statvfs = os.statvfs(ds_rp)

        return round(100 - (100 * statvfs.f_bavail / statvfs.f_blocks), 1)      # may raise FileNotFoundError


# --------------------------------------------------------------------------------------------------------------------

class FilesystemReport(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, path):
        try:
            on_root = Filesystem.is_on_root_filesystem(path)
            used = int(round(Filesystem.percentage_used(path)))

            return cls(path, True, on_root, used)

        except FileNotFoundError:
            return cls(path, False, None, None)


    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        path = jdict.get('path')
        available = jdict.get('is-available')
        on_root = jdict.get('on-root')
        used = jdict.get('used')

        return cls(path, available, on_root, used)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, path, available, on_root, used):
        """
        Constructor
        """
        self.__path = path                          # String
        self.__available = available                # bool
        self.__on_root = on_root                    # bool
        self.__used = used                          # int percentage


    def __eq__(self, other):
        try:
            return self.path == other.path and self.available == other.available and \
                   self.on_root == other.on_root and self.used == other.used

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def path(self):
        return self.__path


    @property
    def available(self):
        return self.__available


    @property
    def on_root(self):
        return self.__on_root


    @property
    def used(self):
        return self.__used


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['path'] = self.path
        jdict['is-available'] = self.available
        jdict['on-root'] = self.on_root
        jdict['used'] = self.used

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "FilesystemReport:{path:%s, available:%s, on_root:%s, used:%s}" % \
               (self.path, self.available, self.on_root, self.used)


# --------------------------------------------------------------------------------------------------------------------

class File(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, container, name):
        abs_filename = os.path.join(container, name)

        return File(container, name, os.path.getmtime(abs_filename), os.path.isdir(abs_filename))


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, container, name, updated, is_directory):
        """
        Constructor
        """
        self.__container = container                            # string
        self.__name = name                                      # string
        self.__updated = int(updated)                           # int timestamp
        self.__is_directory = bool(is_directory)                # bool


    # ----------------------------------------------------------------------------------------------------------------

    def path(self):
        return os.path.join(self.container, self.name)


    def has_suffix(self, suffix):
        return self.name.endswith('.' + suffix)


    def delete(self):
        return Filesystem.rm(self.path())


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def container(self):
        return self.__container


    @property
    def name(self):
        return self.__name


    @property
    def updated(self):
        return self.__updated


    @property
    def is_directory(self):
        return self.__is_directory


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "File:{container:%s, name:%s, updated:%s, is_directory:%s}" %  \
               (self.container, self.name, self.updated, self.is_directory)



"""
Created on 13 Jun 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

Defines the functionality and fields required by PSUMonitor
"""

from abc import ABC, abstractmethod

from scs_core.data.json import JSONReport


# --------------------------------------------------------------------------------------------------------------------

class PSUReport(JSONReport, ABC):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    @abstractmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        pass


    # ----------------------------------------------------------------------------------------------------------------

    # noinspection PyMethodMayBeStatic
    def is_null_datum(self):
        return False


    @abstractmethod
    def below_power_threshold(self, charge_min):
        pass


    # ----------------------------------------------------------------------------------------------------------------

    @property
    @abstractmethod
    def source(self):
        return None


    @property
    @abstractmethod
    def standby(self):
        return None


    @property
    @abstractmethod
    def input_power_present(self):
        return None


    @property
    @abstractmethod
    def v_in(self):
        return None


    @property
    @abstractmethod
    def batt_percent(self):
        return None


    @property
    @abstractmethod
    def charge_status(self):
        return None


    # ----------------------------------------------------------------------------------------------------------------

    @abstractmethod
    def as_json(self):
        pass



"""
Created on 12 Mar 2024

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

Test client for live.southcoastscience.com aggregate finder
"""

import requests

from scs_core.aws.client.api_client import APIClient
from scs_core.aws.config.endpoint import AWSEndpoint
from scs_core.aws.manager.topic_history.topic_history_intercourse import TopicHistoryRequest, TopicHistoryResponse

from scs_core.data.json import JSONify


# --------------------------------------------------------------------------------------------------------------------

class Endpoint(AWSEndpoint):
    @classmethod
    def configuration(cls):
        # https://xy1eszuu23.execute-api.us-west-2.amazonaws.com/dev/topicMessages  - does not work in this context
        # https://xef6bspqs5.execute-api.us-west-2.amazonaws.com/default/AWSAggregate

        return cls('topicMessages',
                   'https://xef6bspqs5.execute-api.us-west-2.amazonaws.com/default/AWSAggregate')


# --------------------------------------------------------------------------------------------------------------------

class AWSAggregateFinder(APIClient):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, reporter=None):
        super().__init__(reporter=reporter)


    # ----------------------------------------------------------------------------------------------------------------

    def find_for_topic(self, topic, start, end, path, fetch_last, checkpoint, include_wrapper, rec_only,
                       min_max, exclude_remainder, fetch_last_written_before, backoff_limit):

        request = TopicHistoryRequest(topic, start, end, path, fetch_last, checkpoint, include_wrapper, rec_only,
                                      min_max, exclude_remainder, fetch_last_written_before, backoff_limit)
        self._logger.debug(request)

        for item in self.__get_blocks(Endpoint.url(), TopicHistoryResponse, params=request.params()):
            yield item


    # ----------------------------------------------------------------------------------------------------------------

    def __get_blocks(self, url, block_class, params=None, payload=None):
        while True:
            response = requests.get(url, headers=self._auth_headers(), params=params, data=JSONify.dumps(payload))
            self._check_response(response)

            self._logger.debug("response: %s" % response.json())

            # messages...
            block = block_class.construct_from_jdict(response.json())
            self._logger.debug("block: %s" % block)

            for item in block.items:
                yield item

            # next request...
            if block.next_url is None:
                break

            params = block.next_params(params)



"""
Created on 27 Sep 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import os
import time

from abc import ABC
from multiprocessing import Process
from subprocess import Popen


# --------------------------------------------------------------------------------------------------------------------

class Network(ABC):
    """
    classdocs
    """
    __TEST_RESOURCE =       'google.com'
    __NETWORK_WAIT_TIME =   5.0                         # seconds


    # ----------------------------------------------------------------------------------------------------------------

    @staticmethod
    def resource_is_available(resource):
        with open(os.devnull, 'wb') as devnull:
            return Popen(['ping', '-c', '1', resource], stdout=devnull, stderr=devnull).wait() == 0


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def is_available(cls):
        return cls.resource_is_available(cls.__TEST_RESOURCE)


    @classmethod
    def wait(cls):
        cls.wait_for_resource(cls.__TEST_RESOURCE)


    @classmethod
    def wait_for_resource(cls, resource):
        while not cls.resource_is_available(resource):
            time.sleep(cls.__NETWORK_WAIT_TIME)


# ----------------------------------------------------------------------------------------------------------------

class NetworkMonitor(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, interval, network_unavailable_handler):
        self.__interval = interval                                                  # float seconds
        self.__network_unavailable_handler = network_unavailable_handler            # function

        self.__proc = None                                                          # Process


    # ----------------------------------------------------------------------------------------------------------------

    def start(self):
        self.__proc = Process(target=self.run)
        self.__proc.start()


    def run(self):
        try:
            Network.wait()

            while True:
                if not Network.is_available():
                    self.__network_unavailable_handler()

                time.sleep(self.__interval)

        except KeyboardInterrupt:
            pass


    def join(self):
        self.__proc.join()


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "NetworkMonitor:{interval:%s, network_unavailable_handler:%s, proc:%s}" % \
               (self.__interval, self.__network_unavailable_handler, self.__proc)




"""
Created on 24 Sep 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

www.alphasense-technology.co.uk...
{"serial_number": null, "type": "DSI", "calibrated_on": "2024-01-01", "dispatched_on": null, "pt1000_v20": null,
"sn1": {"serial_number": "143351067", "sensor_type": "PIDH2", "pid_zero_mv": null, "pid_sensitivity_mv_ppm": 44.9,
"test-calib": {"sensitivity": 0.786, "calibrated-on": "2024-01-05T13:16:04Z"}}}
"""

from collections import OrderedDict

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.datum import Datum
from scs_core.data.json import JSONable

from scs_core.gas.sensor_calib import SensorCalib


# --------------------------------------------------------------------------------------------------------------------

class PIDTestCalib(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        sensitivity = jdict.get('sensitivity')
        calibrated_on = LocalizedDatetime.construct_from_iso8601(jdict.get('calibrated-on'))

        return cls(sensitivity, calibrated_on)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, sensitivity, calibrated_on):
        """
        Constructor
        """
        self.__sensitivity = float(sensitivity)                 # ratio of actual / expected or None
        self.__calibrated_on = calibrated_on                    # LocalizedDatetime


    def __eq__(self, other):
        try:
            return self.sensitivity == other.sensitivity and self.calibrated_on == other.calibrated_on

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['sensitivity'] = self.sensitivity
        jdict['calibrated-on'] = self.calibrated_on

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def sensitivity(self):
        return self.__sensitivity


    @sensitivity.setter
    def sensitivity(self, sensitivity):
        self.__sensitivity = sensitivity


    @property
    def calibrated_on(self):
        return self.__calibrated_on


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "TestCalib:{sensitivity:%s, calibrated_on:%s}" % \
                    (self.sensitivity, self.calibrated_on)


# --------------------------------------------------------------------------------------------------------------------

class PIDCalib(SensorCalib):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        serial_number = jdict.get('serial_number')
        sensor_type = jdict.get('sensor_type')

        if 'pid_zero_mv' in jdict:
            pid_elc_mv = jdict.get('pid_zero_mv')
        else:
            pid_elc_mv = jdict.get('we_sensor_zero_mv')

        if 'pid_sensitivity_mv_ppm' in jdict:
            pid_sens_mv_ppm = jdict.get('pid_sensitivity_mv_ppm')
        else:
            pid_sens_mv_ppm = jdict.get('we_sensitivity_mv_ppb')        # uses mV / ppb field to record mV / ppm

        bump_calib = PIDTestCalib.construct_from_jdict(jdict.get('bump-calib'))

        return cls(serial_number, sensor_type, pid_elc_mv, pid_sens_mv_ppm, bump_calib)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, serial_number, sensor_type, pid_elc_mv, pid_sens_mv_ppm, bump_calib):
        """
        Constructor
        """
        SensorCalib.__init__(self, serial_number, sensor_type)

        self.__pid_elc_mv = Datum.float(pid_elc_mv, 3)                  # PID electronic zero           mV
        self.__pid_sens_mv_ppm = Datum.float(pid_sens_mv_ppm, 6)        # PID sensitivity               mV / ppm
        self.__bump_calib = bump_calib                                  # PIDTestCalib

        self.validate()


    def __eq__(self, other):
        try:
            return self.serial_number == other.serial_number and self.sensor_type == other.sensor_type and \
                   self.pid_elc_mv == other.pid_elc_mv and self.pid_sens_mv_ppm == other.pid_sens_mv_ppm and  \
                   self.bump_calib == other.bump_calib

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def validate(self):
        if self.pid_sens_mv_ppm == 0.0:
            raise ValueError('%s - pid_sensitivity_mv_ppm: zero sensitivity' % self.sensor_type)


    # ----------------------------------------------------------------------------------------------------------------

    def set_defaults(self):
        pass                                # PID may have null values


    def set_sens_mv_from_sens_na(self):
        pass                                # sensitivity nA data is not available


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['serial_number'] = self.serial_number
        jdict['sensor_type'] = self.sensor_type

        jdict['pid_zero_mv'] = self.pid_elc_mv
        jdict['pid_sensitivity_mv_ppm'] = self.pid_sens_mv_ppm
        jdict['bump-calib'] = self.bump_calib

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def pid_elc_mv(self):
        return self.__pid_elc_mv


    @property
    def pid_sens_v_ppb(self):
        return self.__pid_sens_mv_ppm / 1000000.0


    @property
    def pid_sens_mv_ppm(self):
        return self.__pid_sens_mv_ppm


    @property
    def bump_sensitivity(self):
        return None if self.__bump_calib is None else self.__bump_calib.sensitivity


    @property
    def bump_calib(self):
        return self.__bump_calib


    @bump_calib.setter
    def bump_calib(self, bump_calib):
        self.__bump_calib = bump_calib


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "PIDCalib:{serial_number:%s, sensor_type:%s, pid_elc_mv:%s, pid_sens_mv_ppm:%s, bump_calib:%s}" % \
                    (self.serial_number, self.sensor_type, self.pid_elc_mv, self.pid_sens_mv_ppm, self.bump_calib)



"""
Created on 20 Jan 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import json
import re

from collections import OrderedDict

from scs_core.csv.csv_data_log import CSVDataLog, CSVDataLogFile
from scs_core.csv.csv_reader import CSVReader, CSVReaderException

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONable
from scs_core.data.str import Str

from scs_core.sys.filesystem import Filesystem
from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

class CSVLogCursorQueue(JSONable):
    """
    classdocs
    """

    @staticmethod
    def is_data_directory(name):
        return bool(re.fullmatch(r'[1-9]\d{3}-[0-1]\d', name))                  # YYYY-MM


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def find_cursors_for_log(cls, log: CSVDataLog, rec_field):                  # these cursors are NOT live
        cursors = []

        if log.timeline_start is not None:
            for directory_path in cls.__directory_paths(log):                   # may raise FileNotFoundError
                for log_file in cls.__log_files(log, directory_path):
                    cursor = CSVLogCursor.construct_for_log_file(log, log_file, rec_field)

                    if cursor is not None:
                        cursors.append(cursor)

        return cursors


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def __directory_paths(cls, log: CSVDataLog):
        root_directory = Filesystem.ls(log.root_path)
        from_directory = CSVDataLog.directory_name(log.timeline_start)

        if root_directory is None:
            raise FileNotFoundError(log.root_path)

        for item in root_directory:
            if not item.is_directory or not cls.is_data_directory(item.name) or item.name < from_directory:
                continue

            yield item.path()


    @staticmethod
    def __log_files(log: CSVDataLog, directory_path):
        for file in Filesystem.ls(directory_path):
            log_file = CSVDataLogFile.construct(file)

            if log_file is None:
                continue

            if log_file.tag != log.tag:
                continue

            if log_file.topic_subject != log.topic_subject:
                continue

            if log_file.created_datetime.date() < log.timeline_start.date():
                continue

            # TODO: grab current file, now check next file

            yield log_file


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        queue = OrderedDict()

        for cursor_jdict in jdict.get('queue'):
            cursor = CSVLogCursor.construct_from_jdict(cursor_jdict)
            queue[cursor.file_path] = cursor

        return cls(queue)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, queue=None):
        """
        Constructor
        """
        self.__queue = {} if queue is None else queue           # dict of file_path: CSVLogCursor


    def __len__(self):
        return len(self.__queue)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['queue'] = [cursor.as_json() for cursor in self.cursors()]

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def include(self, cursor):
        if cursor is None:
            return

        if cursor.file_path in self.__queue.keys():
            return                                              # assume that the cursor is already live

        if cursor.is_live:
            for key in self.__queue.keys():
                self.__queue[key].is_live = False               # there shall only be one live file

        self.__queue[cursor.file_path] = cursor


    def next(self):
        for cursor in self.cursors():
            return cursor


    def remove(self, file_path):
        try:
            del self.__queue[file_path]
        except KeyError:
            pass


    # ----------------------------------------------------------------------------------------------------------------

    def cursors(self):
        for file_path in sorted(list(self.__queue.keys())):
            yield self.__queue[file_path]


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "CSVLogCursorQueue:{queue:%s}" % Str.collection(self.cursors())


# --------------------------------------------------------------------------------------------------------------------

class CSVLogCursor(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_for_log_file(cls, log: CSVDataLog, log_file, rec_field):          # this cursor is NOT live
        logger = Logging.getLogger()

        reader = None
        row_number = 0

        try:
            reader = CSVReader.construct_for_file(log_file.path(), cast=False)

            for row in reader.rows():
                try:
                    datum = json.loads(row, object_hook=OrderedDict)
                except ValueError:
                    continue

                if rec_field not in datum:
                    raise KeyError(rec_field)

                rec = LocalizedDatetime.construct_from_iso8601(datum[rec_field])

                if rec is None:
                    raise ValueError(datum[rec_field])

                if rec.datetime > log.timeline_start:
                    return cls(log_file.path(), row_number, False)

                row_number += 1

            return None

        except (CSVReaderException, UnicodeDecodeError, ValueError) as ex:
            logger.error("CSVLogCursor: %s: %s" % (log_file.path(), ex))
            return None                                                             # skip corrupt files

        finally:
            if reader is not None:
                reader.close()


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        file_path = jdict.get('file-path')
        row = jdict.get('row')
        is_live = jdict.get('is-live')

        return cls(file_path, row, is_live)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, file_path, row, is_live):
        """
        Constructor
        """
        self.__file_path = file_path                            # string
        self.__row = int(row)                                   # int
        self.__is_live = bool(is_live)                          # bool


    def __eq__(self, other):
        try:
            return self.file_path == other.file_path

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['file-path'] = self.file_path
        jdict['row'] = self.row
        jdict['is-live'] = self.is_live

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def file_path(self):
        return self.__file_path


    @property
    def row(self):
        return self.__row


    @property
    def is_live(self):
        return self.__is_live


    @is_live.setter
    def is_live(self, is_live):
        self.__is_live = is_live


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "CSVLogCursor:{file_path:%s, row:%s, is_live:%s}" %  \
               (self.file_path, self.row, self.is_live)



"""
Created on 6 Dec 2022

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

examples:
ricardo/rural/device/praxis-000431/status
ricardo/rural/loc/1/particulates
"""

from collections import OrderedDict

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONable
from scs_core.data.str import Str


# --------------------------------------------------------------------------------------------------------------------

class TopicPath(JSONable):
    """
    classdocs
   """

    @classmethod
    def is_valid(cls, path_pieces):
        if len(path_pieces) != 5:
            return False

        for piece in path_pieces:
            if len(piece) < 1 or piece.startswith('-') or piece.endswith('-'):
                return False

        if len(path_pieces[0]) < 2:
            return False

        if path_pieces[2] != 'loc' and path_pieces[2] != 'device':
            return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, rec, path):
        path_pieces = path.strip('/').split('/')

        if not cls.is_valid(path_pieces):
            raise ValueError(path)

        return cls(rec, path_pieces)


    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        rec = LocalizedDatetime.construct_from_jdict(jdict.get('rec'))
        path_pieces = jdict.get('path').split('/')

        return cls(rec, path_pieces)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, rec, path_pieces):
        """
        Constructor
        """
        self.__rec = rec                                        # LocalizedDatetime
        self.__path_pieces = path_pieces                        # array of string


    def __lt__(self, other):
        if self.__rec < other.rec:
            return True

        if self.__rec > other.rec:
            return False

        return self.path() < other.path()


    # ----------------------------------------------------------------------------------------------------------------
    # path type...

    def is_device_topic(self):
        return self.__path_pieces[2] == 'device'


    def is_environment_topic(self):
        return self.__path_pieces[2] == 'loc'


    # ----------------------------------------------------------------------------------------------------------------
    # path fields...

    def organisation(self):
        return self.__path_pieces[0]


    def project(self):
        return self.__path_pieces[1]


    def device(self):
        if not self.is_device_topic():
            raise ValueError(self.path())

        return self.__path_pieces[3]


    def location(self):
        if not self.is_environment_topic():
            raise ValueError(self.path())

        return self.__path_pieces[3]


    def phenomenon(self):
        return self.__path_pieces[-1]


    # ----------------------------------------------------------------------------------------------------------------
    # path structure...

    def root(self):
        return self.__path_pieces[0] + '/'


    def generic(self):
        return '/'.join(self.__path_pieces[:-1]) + '/'


    def path(self):
        return '/'.join(self.__path_pieces)


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def rec(self):
        return self.__rec


    @property
    def path_pieces(self):
        return self.__path_pieces


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['rec'] = self.rec
        jdict['path'] = self.path()

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "TopicPath:{rec:%s, path:%s}" % (self.rec, self.path())


# --------------------------------------------------------------------------------------------------------------------

class DeviceTopicGroup(object):
    """
    classdocs
   """

    @classmethod
    def construct(cls, bylines_for_device):
        device_topics = {}
        environment_topics = {}

        for byline in bylines_for_device:
            topic = TopicPath.construct(byline.rec, byline.topic)

            if topic.is_environment_topic():
                environment_topics[topic.generic()] = topic

            elif topic.is_device_topic():
                device_topics[topic.generic()] = topic

        return cls(sorted(device_topics.values()), sorted(environment_topics.values()))


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, device_topics, environment_topics):
        """
        Constructor
        """
        self.__device_topics = device_topics                            # list of TopicPath
        self.__environment_topics = environment_topics                  # list of TopicPath


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def device_topics(self):
        return self.__device_topics


    @property
    def environment_topics(self):
        return self.__environment_topics


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "DeviceTopicGroup:{device_topics:%s, environment_topics:%s}" % \
               (Str.collection(self.device_topics), Str.collection(self.environment_topics))


# --------------------------------------------------------------------------------------------------------------------

class TopicGroup(object):
    """
    classdocs
   """

    @classmethod
    def construct_from_group(cls, topic_byline_group):
        device_topic_groups = {}

        for device in topic_byline_group.devices:
            device_topic_groups[device] = DeviceTopicGroup.construct(topic_byline_group.bylines_for_device(device))

        return cls(device_topic_groups)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, device_topic_groups):
        """
        Constructor
        """
        self.__device_topic_groups = device_topic_groups            # dict of string (device_tag): DeviceTopicGroup


    # ----------------------------------------------------------------------------------------------------------------

    def device_topic_group(self, device):
        return self.__device_topic_groups[device]                   # may raise KeyError


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def device_topic_groups(self):
        return self.__device_topic_groups


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "TopicGroup:{device_topic_groups:%s}" % Str.collection(self.device_topic_groups)



"""
Created on 17 Jun 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import requests

from scs_core.aws.client.api_client import APIClient
from scs_core.aws.config.endpoint import APIEndpoint
from scs_core.aws.monitor.alert.alert_status_intercourse import AlertStatusFindRequest, AlertStatusFindResponse


# --------------------------------------------------------------------------------------------------------------------

class Endpoint(APIEndpoint):
    @classmethod
    def configuration(cls):
        return cls('AlertStatAPI/AlertStatus',
                   'https://n0ctatmvjl.execute-api.us-west-2.amazonaws.com/default/AlertStatus')


# --------------------------------------------------------------------------------------------------------------------

class AlertStatusManager(APIClient):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self):
        super().__init__()


    # ----------------------------------------------------------------------------------------------------------------

    def find(self, token, id_filter, cause_filter, response_mode):
        request = AlertStatusFindRequest(id_filter, cause_filter, response_mode)

        response = requests.get(Endpoint.url(), headers=self._token_headers(token), params=request.params())
        self._check_response(response)

        return AlertStatusFindResponse.construct_from_jdict(response.json())


    def delete(self, token, id):
        url = Endpoint.url(id)

        http_response = requests.delete(url, headers=self._token_headers(token))
        self._check_response(http_response)



"""
Created on 24 Nov 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import requests

from scs_core.aws.client.api_client import APIClient
from scs_core.aws.config.endpoint import APIEndpoint
from scs_core.aws.security.cognito_user import CognitoUserIdentity


# --------------------------------------------------------------------------------------------------------------------

class Endpoint(APIEndpoint):
    @classmethod
    def configuration(cls):
        return cls('CogUsrAPI/CognitoUsersFinder',
                   'https://iw0jza59y1.execute-api.us-west-2.amazonaws.com/default/CognitoUsersFinder')


# --------------------------------------------------------------------------------------------------------------------

class CognitoUserFinder(APIClient):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self):
        super().__init__()


    # ----------------------------------------------------------------------------------------------------------------

    def find_all(self, token):
        url = Endpoint.url('all')

        response = requests.get(url, headers=self._token_headers(token))
        self._check_response(response)

        return tuple(CognitoUserIdentity.construct_from_jdict(jdict) for jdict in response.json())


    def find_by_status(self, token, confirmation_status):
        url = Endpoint.url(confirmation_status.lower())

        response = requests.get(url, headers=self._token_headers(token))
        self._check_response(response)

        return tuple(CognitoUserIdentity.construct_from_jdict(jdict) for jdict in response.json())


    def find_by_enabled(self, token, enabled):
        url = Endpoint.url('enabled' if enabled else 'disabled')

        response = requests.get(url, headers=self._token_headers(token))
        self._check_response(response)

        return tuple(CognitoUserIdentity.construct_from_jdict(jdict) for jdict in response.json())


    def find_by_email(self, token, email):
        url = Endpoint.url('in')
        payload = {"Email": email}

        response = requests.get(url, headers=self._token_headers(token), json=payload)
        self._check_response(response)

        return tuple(CognitoUserIdentity.construct_from_jdict(jdict) for jdict in response.json())


    def get_by_email(self, token, email):
        url = Endpoint.url('exact')
        payload = {"Email": email}

        response = requests.get(url, headers=self._token_headers(token), json=payload)
        self._check_response(response)

        return CognitoUserIdentity.construct_from_jdict(response.json())


    def get_self(self, token):
        url = Endpoint.url('self')

        response = requests.get(url, headers=self._token_headers(token))
        self._check_response(response)

        return CognitoUserIdentity.construct_from_jdict(response.json())
